/*-
# X-BASED CUBES
#
#  Cubes.c
#
###
#
#  Copyright (c) 1994 - 2002	David Albert Bagley, bagleyd@tux.org
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/* Methods file for Cubes */

#include <stdio.h>
#include <stdlib.h>
#ifdef VMS
#include <unixlib.h>
#else
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#endif
#include <X11/IntrinsicP.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/CoreP.h>
#include "CubesP.h"

#ifndef DATAFILE
#define DATAFILE "/usr/games/lib/cubes.dat"
#endif

static void InitializeCubes(Widget request, Widget renew);
static void ExposeCubes(Widget renew, XEvent * event, Region region);
static void ResizeCubes(CubesWidget w);
static void DestroyCubes(Widget old);
static Boolean SetValuesCubes(Widget current, Widget request, Widget renew);

static void QuitCubes(CubesWidget w, XEvent * event, char **args, int nArgs);
static void MoveCubesOut(CubesWidget w, XEvent * event, char **args, int nArgs);
static void MoveCubesTop(CubesWidget w, XEvent * event, char **args, int nArgs);
static void MoveCubesLeft(CubesWidget w, XEvent * event, char **args, int nArgs);
static void MoveCubesIn(CubesWidget w, XEvent * event, char **args, int nArgs);
static void MoveCubesRight(CubesWidget w, XEvent * event, char **args, int nArgs);
static void MoveCubesBottom(CubesWidget w, XEvent * event, char **args, int nArgs);
static void SelectCubes(CubesWidget w, XEvent * event, char **args, int nArgs);
static void ReleaseCubes(CubesWidget w, XEvent * event, char **args, int nArgs);
static void RandomizeCubes(CubesWidget w, XEvent * event, char **args, int nArgs);
static void RandomizeCubesMaybe(CubesWidget w, XEvent * event, char **args, int nArgs);
static void GetCubes(CubesWidget w, XEvent * event, char **args, int nArgs);
static void WriteCubes(CubesWidget w, XEvent * event, char **args, int nArgs);
static void ClearCubes(CubesWidget w, XEvent * event, char **args, int nArgs);
static void UndoCubes(CubesWidget w, XEvent * event, char **args, int nArgs);
static void SolveCubes(CubesWidget w, XEvent * event, char **args, int nArgs);
static void EnterCubes(CubesWidget w, XEvent * event, char **args, int nArgs);
static void LeaveCubes(CubesWidget w, XEvent * event, char **args, int nArgs);
static int  MoveCubes(CubesWidget w, int direction, int control);

static int  PositionToBlock(CubesWidget w, int x, int y, int *i, int *j, int *k);
static int  MovableBlock(CubesWidget w);
static void SelectBlocks(CubesWidget w);
static void SetAllColors(CubesWidget w, Boolean init);
static void CheckBlocks(CubesWidget w);
static void ResetBlocks(CubesWidget w);
static void ResizeBlocks(CubesWidget w);
static void MoveNoBlocks(CubesWidget w);
static int  MoveBlocksDir(CubesWidget w, int direction, Boolean fast);
static void GetBlocks(CubesWidget w);
static void WriteBlocks(CubesWidget w);
static void ClearBlocks(CubesWidget w);
static void UndoBlocks(CubesWidget w);
static void RandomizeBlocks(CubesWidget w);
static void MoveBlocks(CubesWidget w, int from, Boolean fast);
static int  ExchangeBlocks(CubesWidget w, int pos1, int pos2);
static void DrawFrame(CubesWidget w, GC gc);
static void DrawBlock(CubesWidget w, int pos, Boolean blank, Boolean erase,
	int offsetPosition);
static Pixel brighter(CubesWidget w, Pixel pixel);
static Pixel darker(CubesWidget w, Pixel pixel);
static void drawShadow(CubesWidget w, GC gc, int startX, int startY,
	int sizeX, int sizeY);
static void fill3DRect(CubesWidget w, GC gc, GC darkerGC, GC brighterGC,
	int x, int y, int width, int height, Boolean raised);
static void DiscreteMoves(CubesWidget w, int numBlocks, int dir, Boolean fast);
static void AnimateMoves(CubesWidget w, int numBlocks, int dir);
static void AnimateMove(CubesWidget w, int fromPos);
static int  ToPosition(CubesWidget w, int column, int row, int stack);
static int  Column(CubesWidget w, int pos);
static int  Row(CubesWidget w, int pos);
static int  Stack(CubesWidget w, int pos);
static int  int2String(char *buf, int number, int base);

static char defaultTranslationsCubes[] =
"<KeyPress>q: Quit()\n\
   Ctrl<KeyPress>C: Quit()\n\
   <KeyPress>o: MoveOut()\n\
   <KeyPress>KP_Divide: MoveOut()\n\
   <KeyPress>R5: MoveOut()\n\
   <KeyPress>Up: MoveTop()\n\
   <KeyPress>osfUp: MoveTop()\n\
   <KeyPress>KP_Up: MoveTop()\n\
   <KeyPress>KP_8: MoveTop()\n\
   <KeyPress>R8: MoveTop()\n\
   <KeyPress>Left: MoveLeft()\n\
   <KeyPress>osfLeft: MoveLeft()\n\
   <KeyPress>KP_Left: MoveLeft()\n\
   <KeyPress>KP_4: MoveLeft()\n\
   <KeyPress>R10: MoveLeft()\n\
   <KeyPress>i: MoveIn()\n\
   <KeyPress>Begin: MoveIn()\n\
   <KeyPress>KP_5: MoveIn()\n\
   <KeyPress>R11: MoveIn()\n\
   <KeyPress>Right: MoveRight()\n\
   <KeyPress>osfRight: MoveRight()\n\
   <KeyPress>KP_Right: MoveRight()\n\
   <KeyPress>KP_6: MoveRight()\n\
   <KeyPress>R12: MoveRight()\n\
   <KeyPress>Down: MoveBottom()\n\
   <KeyPress>osfDown: MoveBottom()\n\
   <KeyPress>KP_Down: MoveBottom()\n\
   <KeyPress>KP_2: MoveBottom()\n\
   <KeyPress>R14: MoveBottom()\n\
   <Btn1Down>: Select()\n\
   <Btn1Up>: Release()\n\
   <KeyPress>r: Randomize()\n\
   <Btn3Down>(2+): Randomize()\n\
   <Btn3Down>: RandomizeMaybe()\n\
   <KeyPress>g: Get()\n\
   <KeyPress>w: Write()\n\
   <KeyPress>c: Clear()\n\
   <KeyPress>u: Undo()\n\
   <KeyPress>s: Solve()\n\
   <EnterWindow>: Enter()\n\
   <LeaveWindow>: Leave()";

static XtActionsRec actionsListCubes[] =
{
	{(char *) "Quit", (XtActionProc) QuitCubes},
	{(char *) "MoveOut", (XtActionProc) MoveCubesOut},
	{(char *) "MoveTop", (XtActionProc) MoveCubesTop},
	{(char *) "MoveLeft", (XtActionProc) MoveCubesLeft},
	{(char *) "MoveIn", (XtActionProc) MoveCubesIn},
	{(char *) "MoveRight", (XtActionProc) MoveCubesRight},
	{(char *) "MoveBottom", (XtActionProc) MoveCubesBottom},
	{(char *) "Select", (XtActionProc) SelectCubes},
	{(char *) "Release", (XtActionProc) ReleaseCubes},
	{(char *) "Randomize", (XtActionProc) RandomizeCubes},
	{(char *) "RandomizeMaybe", (XtActionProc) RandomizeCubesMaybe},
	{(char *) "Get", (XtActionProc) GetCubes},
	{(char *) "Write", (XtActionProc) WriteCubes},
	{(char *) "Clear", (XtActionProc) ClearCubes},
	{(char *) "Undo", (XtActionProc) UndoCubes},
	{(char *) "Solve", (XtActionProc) SolveCubes},
	{(char *) "Enter", (XtActionProc) EnterCubes},
	{(char *) "Leave", (XtActionProc) LeaveCubes}
};

static XtResource resourcesCubes[] =
{
	{XtNuserName, XtCUserName, XtRString, sizeof (String),
	 XtOffset(CubesWidget, cubes.username),
	 XtRString, (caddr_t) "nobody"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(CubesWidget, cubes.foreground),
	 XtRString, (caddr_t) "Cyan" /*XtDefaultForeground*/},
	{XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(CubesWidget, cubes.background),
	 XtRString, (caddr_t) "NavyBlue" /*XtDefaultBackground*/},
	{XtNblockColor, XtCColor, XtRPixel, sizeof (Pixel),
	 XtOffset(CubesWidget, cubes.blockColor),
	 XtRString, (caddr_t) "LightGray" /*XtDefaultForeground*/},
	{XtNblockBorder, XtCColor, XtRPixel, sizeof (Pixel),
	 XtOffset(CubesWidget, cubes.borderColor),
	 XtRString, (caddr_t) "gray25" /*XtDefaultBackground*/},
	{XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension),
	 XtOffset(CubesWidget, core.width),
	 XtRString, (caddr_t) "100"},
	{XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
	 XtOffset(CubesWidget, core.height),
	 XtRString, (caddr_t) "300"},
	{XtNsizeX, XtCSizeX, XtRInt, sizeof (int),
	 XtOffset(CubesWidget, cubes.sizeX),
	 XtRString, (caddr_t) "4"},	/* DEFAULTCUBESX */
	{XtNsizeY, XtCSizeY, XtRInt, sizeof (int),
	 XtOffset(CubesWidget, cubes.sizeY),
	 XtRString, (caddr_t) "4"},	/* DEFAULTCUBESY */
	{XtNsizeZ, XtCSizeZ, XtRInt, sizeof (int),
	 XtOffset(CubesWidget, cubes.sizeZ),
	 XtRString, (caddr_t) "1"},	/* DEFAULTCUBESZ */
	{XtNmono, XtCMono, XtRBoolean, sizeof (Boolean),
	 XtOffset(CubesWidget, cubes.mono),
	 XtRString, (caddr_t) "FALSE"},
	{XtNreverse, XtCReverse, XtRBoolean, sizeof (Boolean),
	 XtOffset(CubesWidget, cubes.reverse),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmoveDelay, XtCMoveDelay, XtRInt, sizeof (int),
	 XtOffset(CubesWidget, cubes.moveDelay),
	 XtRString, (caddr_t) "10"},
	{XtNbase, XtCBase, XtRInt, sizeof (int),
	 XtOffset(CubesWidget, cubes.base),
	 XtRString, (caddr_t) "10"},
	{XtNstart, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(CubesWidget, cubes.started),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmenu, XtCMenu, XtRInt, sizeof (int),
	 XtOffset(CubesWidget, cubes.menu),
	 XtRString, (caddr_t) "-1"},
	{XtNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
	 XtOffset(CubesWidget, cubes.select), XtRCallback, NULL}
};

CubesClassRec cubesClassRec =
{
	{
		(WidgetClass) & widgetClassRec,		/* superclass */
		(char *) "Cubes",	/* class name */
		sizeof (CubesRec),	/* widget size */
		NULL,		/* class initialize */
		NULL,		/* class part initialize */
		FALSE,		/* class inited */
		(XtInitProc) InitializeCubes,	/* initialize */
		NULL,		/* initialize hook */
		XtInheritRealize,	/* realize */
		actionsListCubes,	/* actions */
		XtNumber(actionsListCubes),	/* num actions */
		resourcesCubes,	/* resources */
		XtNumber(resourcesCubes),	/* num resources */
		NULLQUARK,	/* xrm class */
		TRUE,		/* compress motion */
		TRUE,		/* compress exposure */
		TRUE,		/* compress enterleave */
		TRUE,		/* visible interest */
		(XtWidgetProc) DestroyCubes,	/* destroy */
		(XtWidgetProc) ResizeCubes,	/* resize */
		(XtExposeProc) ExposeCubes,	/* expose */
		(XtSetValuesFunc) SetValuesCubes,	/* set values */
		NULL,		/* set values hook */
		XtInheritSetValuesAlmost,	/* set values almost */
		NULL,		/* get values hook */
		NULL,		/* accept focus */
		XtVersion,	/* version */
		NULL,		/* callback private */
		defaultTranslationsCubes,	/* tm table */
		NULL,		/* query geometry */
		NULL,		/* display accelerator */
		NULL		/* extension */
	},
	{
		0		/* ignore */
	}
};

WidgetClass cubesWidgetClass = (WidgetClass) & cubesClassRec;

#ifndef HAVE_USLEEP
#if !defined( VMS ) || defined( XVMSUTILS ) ||  ( __VMS_VER >= 70000000 )
#ifdef USE_XVMSUTILS
#include <X11/unix_time.h>
#endif
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#if HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#endif
#endif
#if defined(SYSV) || defined(SVR4)
#ifdef LESS_THAN_AIX3_2
#include <sys/poll.h>
#else /* !LESS_THAN_AIX3_2 */
#include <poll.h>
#endif /* !LESS_THAN_AIX3_2 */
#endif /* defined(SYSV) || defined(SVR4) */

/* not static in case usleep found in system include */
int
usleep(unsigned int usec)
{
#if (defined (SYSV) || defined(SVR4)) && !defined(__hpux)
#if defined(HAVE_NANOSLEEP)
	{
		struct timespec rqt;

		rqt.tv_nsec = 1000 * (usec % (unsigned int) 1000000);
		rqt.tv_sec = usec / (unsigned int) 1000000;
		return nanosleep(&rqt, NULL);
	}
#else
	(void) poll(
#if defined(__cplusplus) || defined(c_plusplus)
		(pollfd *) /* guess */
#else
		(void *)
#endif
		0, (int) 0, usec / 1000);	/* ms resolution */
#endif
#else
#ifdef VMS
	long        timadr[2];

	if (usec != 0) {
		timadr[0] = -usec * 10;
		timadr[1] = -1;

		sys$setimr(4, &timadr, 0, 0, 0);
		sys$waitfr(4);
	}
#else
	struct timeval time_out;

#if  0
	/* (!defined(AIXV3) && !defined(__hpux)) */
	extern int  select(int, fd_set *, fd_set *, fd_set *, struct timeval *);

#endif

	time_out.tv_usec = usec % (unsigned int) 1000000;
	time_out.tv_sec = usec / (unsigned int) 1000000;
	(void) select(0, (void *) 0, (void *) 0, (void *) 0, &time_out);
#endif
#endif
	return 0;
}
#endif

static void
Sleep(unsigned int cMilliseconds)
{
	(void) usleep(cMilliseconds * 1000);
}

static void
InitializeCubes(Widget request, Widget renew)
{
	CubesWidget w = (CubesWidget) renew;

	w->cubes.blockOfPosition = NULL;
	CheckBlocks(w);
	InitMoves();
	ResetBlocks(w);
	w->cubes.numSlices = ((w->cubes.moveDelay < MAXSLICES) ?
		w->cubes.moveDelay + 1 : MAXSLICES);
	(void) SRAND(getpid());
	w->cubes.depth = DefaultDepthOfScreen(XtScreen(w));
	SetAllColors(w, True);
	ResizeCubes(w);
}

static void
DestroyCubes(Widget old)
{
	CubesWidget w = (CubesWidget) old;

	XtReleaseGC(old, w->cubes.blockGC);
	XtReleaseGC(old, w->cubes.blockBrighterGC);
	XtReleaseGC(old, w->cubes.blockDarkerGC);
	XtReleaseGC(old, w->cubes.borderGC);
	XtReleaseGC(old, w->cubes.puzzleGC);
	XtReleaseGC(old, w->cubes.inverseGC);
	XtRemoveCallbacks(old, XtNselectCallback, w->cubes.select);
}

static void
ResizeCubes(CubesWidget w)
{
	w->cubes.delta.x = 3;
	w->cubes.delta.y = 3;
	w->cubes.offset.x = MAX(((int) w->core.width - w->cubes.delta.x) /
				w->cubes.sizeX, 0);
	w->cubes.offset.y = MAX(((int) w->core.height - w->cubes.delta.y) /
				w->cubes.sizeY, 0);
	if (w->cubes.offset.y >= w->cubes.offset.x) {
		w->cubes.vertical = TRUE;
		w->cubes.offset.y = MAX(((int) w->core.height / w->cubes.sizeZ -
				      w->cubes.delta.y) / w->cubes.sizeY, 0);
	} else {
		w->cubes.vertical = FALSE;
		w->cubes.offset.x = MAX(((int) w->core.width / w->cubes.sizeZ -
				      w->cubes.delta.x) / w->cubes.sizeX, 0);
	}
	w->cubes.faceSize.x = w->cubes.offset.x * w->cubes.sizeX +
		w->cubes.delta.x + 2;
	w->cubes.faceSize.y = w->cubes.offset.y * w->cubes.sizeY +
		w->cubes.delta.y + 2;
	if (w->cubes.vertical) {
		w->cubes.puzzleSize.x = w->cubes.faceSize.x;
		w->cubes.puzzleSize.y = (w->cubes.faceSize.y - w->cubes.delta.y) *
			w->cubes.sizeZ + w->cubes.delta.y;
	} else {
		w->cubes.puzzleSize.x = (w->cubes.faceSize.x - w->cubes.delta.x) *
			w->cubes.sizeZ + w->cubes.delta.x;
		w->cubes.puzzleSize.y = w->cubes.faceSize.y;
	}
	w->cubes.puzzleOffset.x = ((int) w->core.width -
				   w->cubes.puzzleSize.x + 2) / 2;
	w->cubes.puzzleOffset.y = ((int) w->core.height -
				   w->cubes.puzzleSize.y + 2) / 2;
	w->cubes.blockSize.x = MAX(w->cubes.offset.x - w->cubes.delta.x, 0);
	w->cubes.blockSize.y = MAX(w->cubes.offset.y - w->cubes.delta.y, 0);
	ResizeBlocks(w);
}

static void
ExposeCubes(Widget renew, XEvent * event, Region region)
{
	CubesWidget w = (CubesWidget) renew;

	if (w->core.visible) {
		XFillRectangle(XtDisplay(w), XtWindow(w), w->cubes.inverseGC,
			0, 0, w->core.width, w->core.height);
		DrawFrame(w, (w->cubes.focus) ? w->cubes.puzzleGC :
			w->cubes.inverseGC);
		DrawAllBlocks(w);
	}
}

static      Boolean
SetValuesCubes(Widget current, Widget request, Widget renew)
{
	CubesWidget c = (CubesWidget) current, w = (CubesWidget) renew;
	Boolean     redraw = FALSE;
	Boolean     redrawBlocks = FALSE;

	CheckBlocks(w);
	if (w->cubes.background != c->cubes.background ||
	    w->cubes.foreground != c->cubes.foreground ||
	    w->cubes.borderColor != c->cubes.borderColor ||
	    w->cubes.blockColor != c->cubes.blockColor ||
	    w->cubes.reverse != c->cubes.reverse ||
	    w->cubes.mono != c->cubes.mono) {
		SetAllColors(w, False);
		redrawBlocks = True;
	}
	if (w->cubes.sizeX != c->cubes.sizeX ||
	    w->cubes.sizeY != c->cubes.sizeY ||
	    w->cubes.sizeZ != c->cubes.sizeZ ||
	    w->cubes.base != c->cubes.base) {
		ResetBlocks(w);
		ResizeCubes(w);
		redraw = TRUE;
	} else if (w->cubes.offset.x != c->cubes.offset.x ||
		   w->cubes.offset.y != c->cubes.offset.y) {
		ResizeCubes(w);
		redraw = TRUE;
	}
	if (w->cubes.moveDelay != c->cubes.moveDelay) {
		w->cubes.numSlices = ((w->cubes.moveDelay < MAXSLICES) ?
			w->cubes.moveDelay + 1 : MAXSLICES);
	}
	if (w->cubes.menu != -1) {
		switch (w->cubes.menu) {
		case 0:
			w->cubes.menu = -1;
			GetBlocks(w);
			break;
		case 1:
			w->cubes.menu = -1;
			WriteBlocks(w);
			break;
		case 3:
			w->cubes.menu = -1;
			ClearBlocks(w);
			break;
		case 4:
			w->cubes.menu = -1;
			UndoBlocks(w);
			break;
		case 5:
			w->cubes.menu = -1;
			RandomizeBlocks(w);
			break;
		case 6:
			w->cubes.menu = -1;
#if 0
			SolveBlocks(w);	/* Sorry, this is not implemented */
#endif
			break;
		default:
			w->cubes.menu = -1;
			break;
		}
	}
	if (redrawBlocks && !redraw && XtIsRealized(renew) && renew->core.visible) {
		DrawFrame(c, c->cubes.inverseGC);
		if (w->cubes.focus)
			DrawFrame(w, w->cubes.puzzleGC);
		DrawAllBlocks(w);
	}
	return (redraw);
}

static void
QuitCubes(CubesWidget w, XEvent * event, char **args, int nArgs)
{
	XtCloseDisplay(XtDisplay(w));
	exit(0);
}

static void
SelectCubes(CubesWidget w, XEvent * event, char **args, int nArgs)
{
	int         pos, i, j, k, rowType;

	pos = PositionToBlock(w, event->xbutton.x, event->xbutton.y, &i, &j, &k);
	if (pos >= 0) {
		if (CheckSolved(w)) {
			MoveNoBlocks(w);
			w->cubes.currentPosition = -1;
			return;
		}
		w->cubes.currentPosition = pos;
		w->cubes.currentRow[0] = i;
		w->cubes.currentRow[1] = j;
		w->cubes.currentRow[2] = k;
		rowType = MovableBlock(w);
		if (rowType < 0) {
			cubesCallbackStruct cb;

			DrawBlock(w, w->cubes.currentPosition, rowType == CUBES_SPACE, False, TRUE);
			XFlush(XtDisplay(w));
			Sleep(100);
			DrawBlock(w, w->cubes.currentPosition, True, True, TRUE);
			if (rowType != CUBES_SPACE)
				DrawBlock(w, w->cubes.currentPosition, False, False, FALSE);
			cb.reason = rowType;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
			w->cubes.currentPosition = -1;
			return;
		}
		DrawBlock(w, w->cubes.currentPosition, False, False, TRUE);
	} else
		w->cubes.currentPosition = -1;
}

static void
ReleaseCubes(CubesWidget w, XEvent * event, char **args, int nArgs)
{
	if (w->cubes.currentPosition == -1)
		return;
	DrawBlock(w, w->cubes.currentPosition, True, True, TRUE);
	DrawBlock(w, w->cubes.currentPosition, False, False, FALSE);
	SelectBlocks(w);
	w->cubes.currentPosition = -1;
}

static void
RandomizeCubes(CubesWidget w, XEvent * event, char **args, int nArgs)
{
	RandomizeBlocks(w);
}

static void
RandomizeCubesMaybe(CubesWidget w, XEvent * event, char **args, int nArgs)
{
	if (!w->cubes.started)
		RandomizeBlocks(w);
}

static void
GetCubes(CubesWidget w, XEvent * event, char **args, int nArgs)
{
	GetBlocks(w);
}

static void
WriteCubes(CubesWidget w, XEvent * event, char **args, int nArgs)
{
	WriteBlocks(w);
}

static void
ClearCubes(CubesWidget w, XEvent * event, char **args, int nArgs)
{
	ClearBlocks(w);
}

static void
UndoCubes(CubesWidget w, XEvent * event, char **args, int nArgs)
{
	UndoBlocks(w);
}

static void
SolveCubes(CubesWidget w, XEvent * event, char **args, int nArgs)
{
#if 0
	SolveBlocks(w);		/* Sorry, this is not implemented */
#endif
}

static void
EnterCubes(CubesWidget w, XEvent * event, char **args, int nArgs)
{
	w->cubes.focus = True;
	DrawFrame(w, w->cubes.puzzleGC);
}

static void
LeaveCubes(CubesWidget w, XEvent * event, char **args, int nArgs)
{
	w->cubes.focus = False;
	DrawFrame(w, w->cubes.inverseGC);
}

static void
MoveCubesOut(CubesWidget w, XEvent * event, char **args, int nArgs)
{
	(void) MoveCubes(w, OUT, (int) (event->xkey.state & ControlMask));
}

static void
MoveCubesTop(CubesWidget w, XEvent * event, char **args, int nArgs)
{
	(void) MoveCubes(w, TOP, (int) (event->xkey.state & ControlMask));
}

static void
MoveCubesLeft(CubesWidget w, XEvent * event, char **args, int nArgs)
{
	(void) MoveCubes(w, LEFT, (int) (event->xkey.state & ControlMask));
}

static void
MoveCubesIn(CubesWidget w, XEvent * event, char **args, int nArgs)
{
	(void) MoveCubes(w, IN, (int) (event->xkey.state & ControlMask));
}

static void
MoveCubesRight(CubesWidget w, XEvent * event, char **args, int nArgs)
{
	(void) MoveCubes(w, RIGHT, (int) (event->xkey.state & ControlMask));
}

static void
MoveCubesBottom(CubesWidget w, XEvent * event, char **args, int nArgs)
{
	(void) MoveCubes(w, BOTTOM, (int) (event->xkey.state & ControlMask));
}

static int
MoveCubes(CubesWidget w, int direction, int control)
{
	cubesCallbackStruct cb;

	if (control) {
		cb.reason = CUBES_IGNORE;
		switch (direction) {
			case TOP:
				if (w->cubes.sizeY <= MINCUBES)
					return FALSE;
				cb.reason = CUBES_DEC_Y;
				break;
			case RIGHT:
				cb.reason = CUBES_INC_X;
				break;
			case BOTTOM:
				cb.reason = CUBES_INC_Y;
				break;
			case LEFT:
				if (w->cubes.sizeX <= MINCUBES)
					return FALSE;
				cb.reason = CUBES_DEC_X;
				break;
			case IN:
				if (w->cubes.sizeZ <= MINCUBES)
					return FALSE;
				cb.reason = CUBES_DEC_Z;
				break;
			case OUT:
				cb.reason = CUBES_INC_Z;
				break;
			default:
				(void) printf("MoveCubes: direction %d\n", direction);
		}
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		return FALSE;
	}
	if (CheckSolved(w)) {
		MoveNoBlocks(w);
		return FALSE;
	}
	if (!MoveCubesDir(w, direction, False)) {
		cb.reason = CUBES_BLOCKED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		return FALSE;
	}
	if (CheckSolved(w)) {
		cb.reason = CUBES_SOLVED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
	return TRUE;
}

int
MoveCubesDir(CubesWidget w, int direction, Boolean fast)
{
	cubesCallbackStruct cb;

	if (MoveBlocksDir(w, direction, fast)) {
		cb.reason = CUBES_MOVED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		PutMove(direction);
		return TRUE;
	}
	return FALSE;
}

static int
PositionToBlock(CubesWidget w, int x, int y, int *i, int *j, int *k)
{
	if (w->cubes.vertical) {
		*i = (x - w->cubes.delta.x / 2 -
		      w->cubes.puzzleOffset.x) / w->cubes.offset.x;
		*j = ((y - w->cubes.delta.y / 2 -
		       w->cubes.puzzleOffset.y) % (w->cubes.sizeY * w->cubes.offset.y +
				  w->cubes.delta.y - 1)) / w->cubes.offset.y;
		*k = (y - w->cubes.delta.y / 2 -
		      w->cubes.puzzleOffset.y) / (w->cubes.sizeY * w->cubes.offset.y +
						  w->cubes.delta.y - 1);
	} else {
		*i = ((x - w->cubes.delta.x / 2 -
		       w->cubes.puzzleOffset.x) % (w->cubes.sizeX * w->cubes.offset.x +
				  w->cubes.delta.x - 1)) / w->cubes.offset.x;
		*j = (y - w->cubes.delta.y / 2 -
		      w->cubes.puzzleOffset.y) / w->cubes.offset.y;
		*k = (x - w->cubes.delta.x / 2 -
		      w->cubes.puzzleOffset.x) / (w->cubes.sizeX * w->cubes.offset.x +
						  w->cubes.delta.x - 1);
	}
	if (*i >= 0 && *j >= 0 && *k >= 0 &&
	    *i < w->cubes.sizeX && *j < w->cubes.sizeY && *k < w->cubes.sizeZ)
		return ToPosition(w, *i, *j, *k);
	else
		return -1;
}

static int
MovableBlock(CubesWidget w)
{
	int         i, j, k, l, m;

	i = w->cubes.currentRow[0];
	j = w->cubes.currentRow[1];
	k = w->cubes.currentRow[2];
	l = i + w->cubes.sizeX * j - w->cubes.spacePosition % w->cubes.sizeRect;
	m = i + w->cubes.sizeX * j + w->cubes.sizeRect * k - w->cubes.spacePosition;

	/* Order important if w->cubes.sizeX = 1 */
	if (l % w->cubes.sizeX == 0 && k == Stack(w, w->cubes.spacePosition)) {
		if (l == 0) {
			return CUBES_SPACE;
		}
	} else if (!(l / w->cubes.sizeX == 0 &&
		     j == Row(w, w->cubes.spacePosition) &&
		     k == Stack(w, w->cubes.spacePosition)) &&
		   (m % w->cubes.sizeRect != 0)) {
		return CUBES_BLOCKED;
	}
	return 0;
}

static void
SelectBlocks(CubesWidget w)
{
	cubesCallbackStruct cb;
	int         i, j, k, l, m, n;

	i = w->cubes.currentRow[0];
	j = w->cubes.currentRow[1];
	k = w->cubes.currentRow[2];
	l = i + w->cubes.sizeX * j - w->cubes.spacePosition % w->cubes.sizeRect;
	m = i + w->cubes.sizeX * j + w->cubes.sizeRect * k - w->cubes.spacePosition;

	/* Order important if w->cubes.sizeX = 1 */
	if (l % w->cubes.sizeX == 0 && k == Stack(w, w->cubes.spacePosition)) {
		if (l < 0) {
			if (w->cubes.moveDelay > 0)
				AnimateMoves(w, l / w->cubes.sizeX, BOTTOM);
			else
				for (n = 1; n <= -l / w->cubes.sizeX; n++)
				  DiscreteMoves(w,
				    w->cubes.spacePosition - w->cubes.sizeX,
				    BOTTOM, False);
		} else if (l > 0) {
			if (w->cubes.moveDelay > 0)
				AnimateMoves(w, l / w->cubes.sizeX, TOP);
			else
				for (n = 1; n <= l / w->cubes.sizeX; n++)
				  DiscreteMoves(w,
				    w->cubes.spacePosition + w->cubes.sizeX,
				    TOP, False);
		} else {	/* (l == 0) */
			(void) printf("SelectBlocks: %d\n", CUBES_SPACE);
			return;
		}
	} else if (l / w->cubes.sizeX == 0 &&
		   j == Row(w, w->cubes.spacePosition) &&
		   k == Stack(w, w->cubes.spacePosition)) {
		if (l < 0) {
			if (w->cubes.moveDelay > 0)
				AnimateMoves(w, l, RIGHT);
			else
				for (n = 1; n <= -l % w->cubes.sizeX; n++)
				  DiscreteMoves(w,
				    w->cubes.spacePosition - 1, RIGHT, False);
		} else {	/* (l > 0) */
			if (w->cubes.moveDelay > 0)
				AnimateMoves(w, l, LEFT);
			else
				for (n = 1; n <= l % w->cubes.sizeX; n++)
				  DiscreteMoves(w,
				    w->cubes.spacePosition + 1, LEFT, False);
		}
	} else if (m % w->cubes.sizeRect == 0) {
		if (m < 0) {
			for (n = 1; n <= -Stack(w, m); n++) {
				DiscreteMoves(w,
				  w->cubes.spacePosition - w->cubes.sizeRect,
				  IN, w->cubes.moveDelay <= 0);
			}
		} else {	/* (m > 0) */
			for (n = 1; n <= Stack(w, m); n++) {
				DiscreteMoves(w,
				  w->cubes.spacePosition + w->cubes.sizeRect,
				  OUT, w->cubes.moveDelay <= 0);
			}
		}
	} else {
		(void) printf("SelectBlocks: %d\n", CUBES_BLOCKED);
		return;
	}
	if (CheckSolved(w)) {
		cb.reason = CUBES_SOLVED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
}

#define FACTOR 0.7
#define MAXINTENSITY 0xFFFF

static Pixel
brighter(CubesWidget w, Pixel pixel)
{
	XColor color;
	int i = (int) ((1 - FACTOR) * MAXINTENSITY);

	color.pixel = pixel;
	XQueryColor(XtDisplay(w), DefaultColormap(XtDisplay(w), XtWindow(w)),
		&color);
	if (color.red < i)
		color.red = i;
	if (color.green < i)
		color.green = i;
	if (color.blue < i)
		color.blue = i;
	color.red = (unsigned short) MIN(color.red / FACTOR, MAXINTENSITY);
	color.green = (unsigned short) MIN(color.green / FACTOR, MAXINTENSITY);
	color.blue = (unsigned short) MIN(color.blue / FACTOR, MAXINTENSITY);
	if (XAllocColor(XtDisplay(w),
			DefaultColormap(XtDisplay(w), XtWindow(w)),
			&color))
		return color.pixel;
	return pixel;
}

static Pixel
darker(CubesWidget w, Pixel pixel)
{
	XColor color;

	color.pixel = pixel;
	XQueryColor(XtDisplay(w), DefaultColormap(XtDisplay(w), XtWindow(w)),
		&color);
	color.red = (unsigned short) (color.red * FACTOR);
	color.green = (unsigned short) (color.green * FACTOR);
	color.blue = (unsigned short) (color.blue * FACTOR);
	if (XAllocColor(XtDisplay(w),
			DefaultColormap(XtDisplay(w), XtWindow(w)),
			&color))
		return color.pixel;
	return pixel;
}

static void
SetAllColors(CubesWidget w, Boolean init)
{
	XGCValues   values;
	XtGCMask    valueMask;

	valueMask = GCForeground | GCBackground;

	if (w->cubes.reverse) {
		values.foreground = w->cubes.foreground;
		values.background = w->cubes.background;
	} else {
		values.foreground = w->cubes.background;
		values.background = w->cubes.foreground;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->cubes.inverseGC);
	w->cubes.inverseGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->cubes.reverse) {
		values.foreground = w->cubes.background;
		values.background = w->cubes.foreground;
	} else {
		values.foreground = w->cubes.foreground;
		values.background = w->cubes.background;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->cubes.puzzleGC);
	w->cubes.puzzleGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->cubes.depth < 2 || w->cubes.mono) {
		if (w->cubes.reverse) {
			values.foreground = w->cubes.background;
			values.background = w->cubes.foreground;
		} else {
			values.foreground = w->cubes.foreground;
			values.background = w->cubes.background;
		}
	} else {
		values.foreground = w->cubes.blockColor;
		values.background = w->cubes.borderColor;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->cubes.blockGC);
	w->cubes.blockGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->cubes.depth >= 2 && !w->cubes.mono) {
		values.foreground = brighter(w, w->cubes.blockColor);
	}
	if (!init)
		XtReleaseGC((Widget) w, w->cubes.blockBrighterGC);
	w->cubes.blockBrighterGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->cubes.depth >= 2 && !w->cubes.mono) {
		values.foreground = darker(w, w->cubes.blockColor);
	}
	if (!init)
		XtReleaseGC((Widget) w, w->cubes.blockDarkerGC);
	w->cubes.blockDarkerGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->cubes.depth < 2 || w->cubes.mono) {
		if (w->cubes.reverse) {
			values.foreground = w->cubes.foreground;
			values.background = w->cubes.background;
		} else {
			values.foreground = w->cubes.background;
			values.background = w->cubes.foreground;
		}
	} else {
		values.foreground = w->cubes.borderColor;
		values.background = w->cubes.blockColor;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->cubes.borderGC);
	w->cubes.borderGC = XtGetGC((Widget) w, valueMask, &values);
}

static void
CheckBlocks(CubesWidget w)
{
	char        buf[121];

	if (w->cubes.sizeX < MINCUBES) {
		(void) sprintf(buf,
			"Number of Cubes in X direction incorrect, must be at least %d, defaulting to %d",
			MINCUBES, DEFAULTCUBESX);
		XtWarning(buf);
		w->cubes.sizeX = DEFAULTCUBESX;
	}
	if (w->cubes.sizeY < MINCUBES) {
		(void) sprintf(buf,
			"Number of Cubes in Y direction incorrect, must be at least %d, defaulting to %d",
			MINCUBES, DEFAULTCUBESY);
		XtWarning(buf);
		w->cubes.sizeY = DEFAULTCUBESY;
	}
	if (w->cubes.sizeZ < MINCUBES) {
		(void) sprintf(buf,
			"Number of Cubes in Z direction incorrect, must be at least %d, defaulting to %d",
			MINCUBES, DEFAULTCUBESZ);
		XtWarning(buf);
		w->cubes.sizeZ = DEFAULTCUBESZ;
	}
	if (w->cubes.moveDelay < 0) {
		(void) sprintf(buf,
			"%d delay can not be negative, taking absolute value",
			w->cubes.moveDelay);
		XtWarning(buf);
		w->cubes.moveDelay = -w->cubes.moveDelay;
	}
	if (w->cubes.base > 36) {
		/* 10 numbers + 26 letters (ASCII or EBCDIC) */
		XtWarning("Base must be less than or equal to 36");
		w->cubes.base = 10;
	} else if (w->cubes.base <= 1) {	/* Base 1 is rediculous :) */
		XtWarning("Base must be greater than 1");
		w->cubes.base = 10;
	}
}

static void
ResetBlocks(CubesWidget w)
{
	int         i;

	w->cubes.sizeRect = w->cubes.sizeX * w->cubes.sizeY;
	w->cubes.sizeBlock = w->cubes.sizeRect * w->cubes.sizeZ;
	if (w->cubes.blockOfPosition)
		(void) free((void *) w->cubes.blockOfPosition);
	if (!(w->cubes.blockOfPosition = (int *)
	      malloc(sizeof (int) * w->cubes.sizeBlock)))
		            XtError("Not enough memory, exiting.");

	if (startPosition)
		(void) free((void *) startPosition);
	if (!(startPosition = (int *)
	      malloc(sizeof (int) * w->cubes.sizeBlock)))
		            XtError("Not enough memory, exiting.");

	w->cubes.spacePosition = w->cubes.sizeBlock - 1;
	w->cubes.blockOfPosition[w->cubes.sizeBlock - 1] = 0;
	for (i = 1; i < w->cubes.sizeBlock; i++)
		w->cubes.blockOfPosition[i - 1] = i;
	FlushMoves(w);
	w->cubes.currentPosition = -1;
	w->cubes.started = FALSE;
}

static void
ResizeBlocks(CubesWidget w)
{
	w->cubes.digitOffset.x = 3;
	w->cubes.digitOffset.y = 4;
}

static void
MoveNoBlocks(CubesWidget w)
{
	cubesCallbackStruct cb;

	cb.reason = CUBES_IGNORE;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static int
MoveBlocksDir(CubesWidget w, int direction, Boolean fast)
{
	switch (direction) {
		case TOP:
			if (Row(w, w->cubes.spacePosition) < w->cubes.sizeY - 1) {
				MoveBlocks(w, w->cubes.spacePosition +
					w->cubes.sizeX, fast);
				return TRUE;
			}
			break;
		case RIGHT:
			if (Column(w, w->cubes.spacePosition) > 0) {
				MoveBlocks(w, w->cubes.spacePosition - 1, fast);
				return TRUE;
			}
			break;
		case BOTTOM:
			if (Row(w, w->cubes.spacePosition) > 0) {
				MoveBlocks(w, w->cubes.spacePosition -
					w->cubes.sizeX, fast);
				return TRUE;
			}
			break;
		case LEFT:
			if (Column(w, w->cubes.spacePosition) < w->cubes.sizeX - 1) {
				MoveBlocks(w, w->cubes.spacePosition + 1, fast);
				return TRUE;
			}
			break;
		case IN:
			if (Stack(w, w->cubes.spacePosition) > 0) {
				MoveBlocks(w, w->cubes.spacePosition -
					w->cubes.sizeRect, fast);
				return TRUE;
			}
			break;
		case OUT:
			if (Stack(w, w->cubes.spacePosition) < w->cubes.sizeZ - 1) {
				MoveBlocks(w, w->cubes.spacePosition +
					w->cubes.sizeRect, fast);
				return TRUE;
			}
			break;
		default:
			(void) printf("MoveBlocksDir: direction %d\n", direction);
	}
	return FALSE;
}

static void
GetBlocks(CubesWidget w)
{
	FILE       *fp;
	int         c, i, sizeX, sizeY, sizeZ, moves;
	cubesCallbackStruct cb;

	if ((fp = fopen(DATAFILE, "r")) == NULL)
		(void) printf("Can not read %s for get.\n", DATAFILE);
	else {
		FlushMoves(w);
		while ((c = getc(fp)) != EOF && c != SYMBOL);
		(void) fscanf(fp, "%d", &sizeX);
		if (sizeX >= MINCUBES) {
			for (i = w->cubes.sizeX; i < sizeX; i++) {
				cb.reason = CUBES_INC_X;
				XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
			}
			for (i = w->cubes.sizeX; i > sizeX; i--) {
				cb.reason = CUBES_DEC_X;
				XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
			}
		} else
			(void) printf("%s corrupted: sizeX %d should be between %d and MAXINT\n",
				      DATAFILE, sizeX, MINCUBES);
		while ((c = getc(fp)) != EOF && c != SYMBOL);
		(void) fscanf(fp, "%d", &sizeY);
		if (sizeY >= MINCUBES) {
			for (i = w->cubes.sizeY; i < sizeY; i++) {
				cb.reason = CUBES_INC_Y;
				XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
			}
			for (i = w->cubes.sizeY; i > sizeY; i--) {
				cb.reason = CUBES_DEC_Y;
				XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
			}
		} else
			(void) printf("%s corrupted: sizeY %d should be between %d and MAXINT\n",
				      DATAFILE, sizeY, MINCUBES);
		while ((c = getc(fp)) != EOF && c != SYMBOL);
		(void) fscanf(fp, "%d", &sizeZ);
		if (sizeZ >= MINCUBES) {
			for (i = w->cubes.sizeZ; i < sizeZ; i++) {
				cb.reason = CUBES_INC_Z;
				XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
			}
			for (i = w->cubes.sizeZ; i > sizeZ; i--) {
				cb.reason = CUBES_DEC_Z;
				XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
			}
		} else
			(void) printf("%s corrupted: sizeZ %d should be between %d and MAXINT\n",
				      DATAFILE, sizeZ, MINCUBES);
		while ((c = getc(fp)) != EOF && c != SYMBOL);
		(void) fscanf(fp, "%d", &moves);
		ScanStartPosition(fp, w);
		cb.reason = CUBES_RESTORE;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		SetStartPosition(w);
		ScanMoves(fp, w, moves);
		(void) fclose(fp);
		(void) printf("%s: sizeX %d, sizeY %d, sizeZ %d, moves %d.\n",
			      DATAFILE, sizeX, sizeY, sizeZ, moves);
	}
}

static void
WriteBlocks(CubesWidget w)
{
	FILE       *fp;

	if ((fp = fopen(DATAFILE, "w")) == NULL)
		(void) printf("Can not write to %s.\n", DATAFILE);
	else {
		(void) fprintf(fp, "sizeX%c %d\n", SYMBOL, w->cubes.sizeX);
		(void) fprintf(fp, "sizeY%c %d\n", SYMBOL, w->cubes.sizeY);
		(void) fprintf(fp, "sizeZ%c %d\n", SYMBOL, w->cubes.sizeZ);
		(void) fprintf(fp, "moves%c %d\n", SYMBOL, NumMoves());
		PrintStartPosition(fp, w);
		PrintMoves(fp);
		(void) fclose(fp);
		(void) printf("Saved to %s.\n", DATAFILE);
	}
}

static void
ClearBlocks(CubesWidget w)
{
	cubesCallbackStruct cb;

	ResetBlocks(w);
	DrawAllBlocks(w);
	cb.reason = CUBES_RESET;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static void
UndoBlocks(CubesWidget w)
{
	if (MadeMoves()) {
		int         direction;

		GetMove(&direction);
		direction = (direction < 4) ? (direction + 2) % 4 :
			((direction == IN) ? OUT : IN);

		if (MoveBlocksDir(w, direction, True)) {
			cubesCallbackStruct cb;

			cb.reason = CUBES_UNDO;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		}
	}
}

static void
RandomizeBlocks(CubesWidget w)
{
	cubesCallbackStruct cb;

	/* First interchange blocks an even number of times */
	/* Must be at least 3x2 or 2x2x2 */
	if ((w->cubes.sizeX > 2 && (w->cubes.sizeY > 1 || w->cubes.sizeZ > 1)) ||
			(w->cubes.sizeY > 2 && (w->cubes.sizeX > 1 || w->cubes.sizeZ > 1)) ||
			(w->cubes.sizeZ > 2 && (w->cubes.sizeX > 1 || w->cubes.sizeY > 1)) ||
			(w->cubes.sizeX == 2 && w->cubes.sizeY == 2 && w->cubes.sizeZ == 2)) {
		int         pos, count = 0;

		for (pos = 0; pos < w->cubes.sizeBlock; pos++) {
			int         randomPos = pos;

			while (randomPos == pos)
				randomPos = NRAND(w->cubes.sizeBlock);
			count += ExchangeBlocks(w, pos, randomPos);
		}
		if (count % 2 &&
		    !ExchangeBlocks(w, 0, 1) &&
		    !ExchangeBlocks(w, w->cubes.sizeBlock - 2,
		             w->cubes.sizeBlock - 1))
			(void) printf("RandomizeBlocks: should not get here\n");
		DrawAllBlocks(w);
		XFlush(XtDisplay(w));
	}
	/* Now move the space around randomly */
	if (w->cubes.sizeX > 1 || w->cubes.sizeY > 1 || w->cubes.sizeZ > 1) {
		int         big = w->cubes.sizeBlock + NRAND(2);
		int         lastDirection = -1;
		int         randomDirection;

		cb.reason = CUBES_RESET;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);

#ifdef DEBUG
		big = 3;
#endif

		if (big > 1000)
			big = 1000;
		while (big--) {
			randomDirection = NRAND(COORD);

#ifdef DEBUG
			Sleep(1);
#endif

			if ((randomDirection < 4 && (randomDirection + 2) % 4 != lastDirection) ||
			    (randomDirection >= 4 && randomDirection + ((randomDirection % 2) ? -1 : 1) != lastDirection) ||
			    (w->cubes.sizeX == 1 && w->cubes.sizeY == 1) ||
			    (w->cubes.sizeY == 1 && w->cubes.sizeZ == 1) ||
			    (w->cubes.sizeZ == 1 && w->cubes.sizeX == 1)) {
				if (MoveCubesDir(w, randomDirection, True))
					lastDirection = randomDirection;
				else
					big++;
			}
		}
		FlushMoves(w);
		cb.reason = CUBES_RANDOMIZE;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
	if (CheckSolved(w)) {
		cb.reason = CUBES_SOLVED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
}

static void
MoveBlocks(CubesWidget w, int from, Boolean fast)
{
	int         tempBlock;

	tempBlock = w->cubes.blockOfPosition[from];
	if (!fast && w->cubes.moveDelay > 0) {
		int i = w->cubes.currentRow[0] = Column(w, from);
		int j = w->cubes.currentRow[1] = Row(w, from);
		int k = w->cubes.currentRow[2] = Stack(w, from);

		int l = i + w->cubes.sizeX * j - w->cubes.spacePosition % w->cubes.sizeRect;
		/* int m = i + w->cubes.sizeX * j + w->cubes.sizeRect * k - w->cubes.spacePosition; */
		/* Order important if w->cubes.sizeX = 1 */
		if (l % w->cubes.sizeX == 0 && k == Stack(w, w->cubes.spacePosition)) {
			if (l < 0) {
				AnimateMoves(w, l / w->cubes.sizeX, BOTTOM);
			} else if (l > 0) {
				AnimateMoves(w, l / w->cubes.sizeX, TOP);
			}
		} else if (l / w->cubes.sizeX == 0 &&
			   j == Row(w, w->cubes.spacePosition) &&
			   k == Stack(w, w->cubes.spacePosition)) {
			if (l < 0) {
				AnimateMoves(w, l, RIGHT);
			} else {	/* (l > 0) */
				AnimateMoves(w, l, LEFT);
			}
		} else {
			w->cubes.blockOfPosition[from] =
				w->cubes.blockOfPosition[w->cubes.spacePosition];
			w->cubes.blockOfPosition[w->cubes.spacePosition] = tempBlock;
			AnimateMove(w, from);
		}
	} else {
		w->cubes.blockOfPosition[from] =
			w->cubes.blockOfPosition[w->cubes.spacePosition];
		w->cubes.blockOfPosition[w->cubes.spacePosition] = tempBlock;
		DrawBlock(w, w->cubes.spacePosition, False, False, FALSE);
		w->cubes.spacePosition = from;
		DrawBlock(w, w->cubes.spacePosition, True, True, FALSE);
	}
}

static int
ExchangeBlocks(CubesWidget w, int pos1, int pos2)
{
	int         tempBlock;

	if (w->cubes.blockOfPosition[pos1] <= 0)
		return FALSE;
	else if (w->cubes.blockOfPosition[pos2] <= 0)
		return FALSE;
	tempBlock = w->cubes.blockOfPosition[pos1];
	w->cubes.blockOfPosition[pos1] = w->cubes.blockOfPosition[pos2];
	w->cubes.blockOfPosition[pos2] = tempBlock;
	return TRUE;
}

static void
DrawFrame(CubesWidget w, GC gc)
{
	int         sumX, sumY, sumZ, offsetX, offsetY, dz, k;

	offsetX = w->cubes.puzzleOffset.x;
	offsetY = w->cubes.puzzleOffset.y;
	XFillRectangle(XtDisplay(w), XtWindow(w), w->cubes.borderGC,
		0, 0, w->core.width, offsetY);
	XFillRectangle(XtDisplay(w), XtWindow(w), w->cubes.borderGC,
		0, w->core.height - offsetY - 1,
		w->core.width, offsetY + 1);
	XFillRectangle(XtDisplay(w), XtWindow(w), w->cubes.borderGC,
		0, 0, offsetX, w->core.height);
	XFillRectangle(XtDisplay(w), XtWindow(w), w->cubes.borderGC,
		w->core.width - offsetX - 1, 0,
		offsetX + 1, w->core.height);

	sumX = w->cubes.sizeX * w->cubes.offset.x + w->cubes.delta.x / 2 + 1;
	sumY = w->cubes.sizeY * w->cubes.offset.y + w->cubes.delta.y / 2 + 1;
	if (w->cubes.vertical) {
		sumZ = offsetY;
		dz = sumY;
	} else {
		sumZ = offsetX;
		dz = sumX;
	}
	for (k = 0; k < w->cubes.sizeZ; k++) {
		if (w->cubes.vertical) {
			XFillRectangle(XtDisplay(w), XtWindow(w), gc,
				       offsetX, sumZ, 1, sumY);
			XFillRectangle(XtDisplay(w), XtWindow(w), gc,
				       offsetX, sumZ, sumX, 1);
			XFillRectangle(XtDisplay(w), XtWindow(w), gc,
				       sumX + offsetX, sumZ, 1, sumY + 1);
			XFillRectangle(XtDisplay(w), XtWindow(w), gc,
				       offsetX, sumY + sumZ, sumX + 1, 1);
		} else {
			XFillRectangle(XtDisplay(w), XtWindow(w), gc,
				       sumZ, offsetY, 1, sumY);
			XFillRectangle(XtDisplay(w), XtWindow(w), gc,
				       sumZ, offsetY, sumX, 1);
			XFillRectangle(XtDisplay(w), XtWindow(w), gc,
				       sumX + sumZ, offsetY, 1, sumY + 1);
			XFillRectangle(XtDisplay(w), XtWindow(w), gc,
				       sumZ, sumY + offsetY, sumX + 1, 1);
		}
		sumZ += dz;
	}
}

void
DrawAllBlocks(CubesWidget w)
{
	int         k;

	for (k = 0; k < w->cubes.sizeBlock; k++)
		DrawBlock(w, k, (w->cubes.blockOfPosition[k] <= 0), (w->cubes.blockOfPosition[k] <= 0), FALSE);
}

static void
drawShadow(CubesWidget w, GC gc, int startX, int startY,
		int sizeX, int sizeY)
{
	XFillRectangle(XtDisplay(w), XtWindow(w), gc,
		       startX, startY, sizeX + 1, 1);
	XFillRectangle(XtDisplay(w), XtWindow(w), gc,
		       startX, startY, 1, sizeY + 1);
}

static void
fill3DRect(CubesWidget w, GC gc, GC darkerGC, GC brighterGC,
		int x, int y, int width, int height, Boolean raised)
{
	GC currentGC = (raised) ? gc : darkerGC;

	if (width > 2 && height > 2)
		XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
			x + 1, y + 1, width - 2, height - 2);
	currentGC = (raised) ? brighterGC : darkerGC;
	XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
		x, y, 1, height);
	if (width > 1)
		XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
			x + 1, y, width - 1, 1);
	XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
		x + 1, y, 1, height);
	if (width > 2)
		XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
			x + 2, y + 1, width - 2, 1);
	currentGC = (raised) ? darkerGC : gc;
	if (width > 1 && height > 1)
		XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
			x + 1, y + height - 1, width - 1, 1);
	if (width > 1 && height > 1)
		XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
			x + width - 1, y, 1, height - 1);
	if (width > 3 && height > 2)
		XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
			x + 2, y + height - 2, width - 3, 1);
	if (width > 2 && height > 3)
		XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
			x + width - 2, y + 1, 1, height - 3);
}

static void
DrawBlock(CubesWidget w, int pos, Boolean blank, Boolean erase,
		int offsetPosition)
{
	int         dx, dy, dz;
	GC          blockGC, borderGC;

	if (erase) {
		blockGC = w->cubes.inverseGC;
		borderGC = w->cubes.inverseGC;
	} else {
		blockGC = w->cubes.blockGC;
		borderGC = w->cubes.borderGC;
	}
	if (w->cubes.vertical) {
		dx = Column(w, pos) * w->cubes.offset.x + w->cubes.delta.x +
			w->cubes.puzzleOffset.x + offsetPosition;
		dy = Row(w, pos) * w->cubes.offset.y + w->cubes.delta.y +
			w->cubes.puzzleOffset.y + offsetPosition;
		dz = Stack(w, pos) * (w->cubes.sizeY * w->cubes.offset.y +
				      w->cubes.delta.y - 1);
		dy += dz;
	} else {
		dx = Column(w, pos) * w->cubes.offset.x + w->cubes.delta.x +
			w->cubes.puzzleOffset.x + offsetPosition;
		dy = Row(w, pos) * w->cubes.offset.y + w->cubes.delta.y +
			w->cubes.puzzleOffset.y + offsetPosition;
		dz = Stack(w, pos) * (w->cubes.sizeX * w->cubes.offset.x +
				      w->cubes.delta.x - 1);
		dx += dz;
	}
	if (blank) {
		XFillRectangle(XtDisplay(w), XtWindow(w), blockGC, dx, dy,
		       w->cubes.blockSize.x, w->cubes.blockSize.y);
		XDrawRectangle(XtDisplay(w), XtWindow(w), borderGC, dx, dy,
		       w->cubes.blockSize.x, w->cubes.blockSize.y);
	} else {
		if (offsetPosition != 0) {
			drawShadow(w, w->cubes.blockDarkerGC,
				dx - offsetPosition, dy - offsetPosition,
				w->cubes.blockSize.x - 1,
				w->cubes.blockSize.y - 1);
		}
		fill3DRect(w, blockGC,
			w->cubes.blockDarkerGC, w->cubes.blockBrighterGC,
			dx, dy, w->cubes.blockSize.x, w->cubes.blockSize.y,
			offsetPosition == 0);
	}
	if (!blank) {
		int         i = 0, offsetX = 0, block = w->cubes.blockOfPosition[pos];
		char        buf[5];

		(void) int2String(buf, block, w->cubes.base);
		while (block >= 1) {
			block /= w->cubes.base;
			offsetX += w->cubes.digitOffset.x;
			i++;
		}
		(void) XDrawString(XtDisplay(w), XtWindow(w), borderGC,
			    dx + w->cubes.blockSize.x / 2 - offsetX,
			    dy + w->cubes.blockSize.y / 2 + w->cubes.digitOffset.y, buf, i);
	}
}

static void
DiscreteMoves(CubesWidget w, int numBlocks, int dir, Boolean fast)
{
	cubesCallbackStruct cb;

	MoveBlocks(w, numBlocks, fast);
	cb.reason = CUBES_MOVED;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	PutMove(dir);
}

static void
AnimateMoves(CubesWidget w, int numBlocks, int dir)
{
	cubesCallbackStruct cb;
	int offsetPosition = 0;
	int toMoveX = w->cubes.offset.x;
	int gapX = toMoveX - w->cubes.blockSize.x;
	int toMoveY = w->cubes.offset.y;
	int gapY = toMoveY - w->cubes.blockSize.y;
	int inc, aBlock;
	int fillBeginPos, fillBeginI;

	if ((dir == RIGHT) || (dir == LEFT)) {
		fillBeginPos = w->cubes.spacePosition + numBlocks;
		fillBeginI = Column(w, fillBeginPos) * w->cubes.offset.x +
		  w->cubes.delta.x + w->cubes.puzzleOffset.x + offsetPosition;
		if (numBlocks < 0)
		  numBlocks = -numBlocks;
		for (inc = 0; inc <= toMoveX;) {
		  for (aBlock = 0; aBlock < numBlocks; aBlock++) {
		    int dx, dy, dz, pos;

		    if (inc == 0) {
		      cb.reason = CUBES_MOVED;
		      XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		      PutMove(dir);
		    }
		    if (dir == RIGHT)
		      pos = w->cubes.spacePosition - aBlock - 1;
		    else /* dir == LEFT */
		      pos = w->cubes.spacePosition + aBlock + 1;
		    if (w->cubes.vertical) {
		      dx = Column(w, pos) * w->cubes.offset.x +
		        w->cubes.delta.x + w->cubes.puzzleOffset.x +
		        offsetPosition;
		      dy = Row(w, pos) * w->cubes.offset.y +
		        w->cubes.delta.y + w->cubes.puzzleOffset.y +
		        offsetPosition;
		      dz = Stack(w, pos) * (w->cubes.sizeY *
		        w->cubes.offset.y + w->cubes.delta.y - 1);
		      dy += dz;
		    } else {
		      dx = Column(w, pos) * w->cubes.offset.x +
		        w->cubes.delta.x + w->cubes.puzzleOffset.x +
		        offsetPosition;
		      dy = Row(w, pos) * w->cubes.offset.y +
		        w->cubes.delta.y + w->cubes.puzzleOffset.y +
		        offsetPosition;
		      dz = Stack(w, pos) * (w->cubes.sizeX *
		        w->cubes.offset.x + w->cubes.delta.x - 1);
		      dx += dz;
		      /* just add the first time */
		      if (aBlock == 0 && inc == 0)
		        fillBeginI += dz;
		    }
		    {
		      int ix, iy;
		      iy = dy;
		      ix = dx + ((dir == RIGHT) ? inc : -inc);
		      fill3DRect(w, w->cubes.blockGC,
		        w->cubes.blockDarkerGC, w->cubes.blockBrighterGC,
		        ix, iy, w->cubes.blockSize.x, w->cubes.blockSize.y,
		        True);
		      if (aBlock < numBlocks - 1) {
		        if (dir == RIGHT)
		          XFillRectangle(XtDisplay(w), XtWindow(w),
		            w->cubes.inverseGC,
		            ix - gapX, iy, gapX, w->cubes.blockSize.y);
		        else /* dir == LEFT */
		          XFillRectangle(XtDisplay(w), XtWindow(w),
		            w->cubes.inverseGC,
		            ix + w->cubes.blockSize.x, iy,
		            gapX, w->cubes.blockSize.y);
		      } else {
		        if (dir == RIGHT)
		          XFillRectangle(XtDisplay(w), XtWindow(w),
		            w->cubes.inverseGC,
		            fillBeginI, iy,
		            ix - fillBeginI, w->cubes.blockSize.y);
		        else
		          XFillRectangle(XtDisplay(w), XtWindow(w),
		            w->cubes.inverseGC,
		            ix + w->cubes.blockSize.x, iy,
		            fillBeginI - ix, w->cubes.blockSize.y);
		      }
		      {
		        /* draw block number after slide complete */
		        int i = 0, offsetX = 0;
		        int block = w->cubes.blockOfPosition[pos];
		        char buf[5];

		        (void) int2String(buf, block, w->cubes.base);
		        while (block >= 1) {
		          block /= w->cubes.base;
		          offsetX += w->cubes.digitOffset.x;
		          i++;
		        }
		        (void) XDrawString(XtDisplay(w), XtWindow(w),
		          w->cubes.borderGC,
			  ix + w->cubes.blockSize.x / 2 - offsetX,
			  iy + w->cubes.blockSize.y / 2 +
			  w->cubes.digitOffset.y, buf, i);
		      }
		    }
		  }
		  if (inc == toMoveX)
		    break;
		  if (toMoveX - inc > toMoveX / w->cubes.numSlices)
		    inc += toMoveX / w->cubes.numSlices;
		  else
		    inc = toMoveX;
		  XFlush(XtDisplay(w));
		  Sleep(w->cubes.moveDelay);
		}
		if (dir == RIGHT)
		  for (aBlock = 0; aBlock < numBlocks; aBlock++)
		    w->cubes.blockOfPosition[w->cubes.spacePosition - aBlock] =
		      w->cubes.blockOfPosition[w->cubes.spacePosition - aBlock - 1];
		else
		  for (aBlock = 0; aBlock < numBlocks; aBlock++)
	 	    w->cubes.blockOfPosition[w->cubes.spacePosition + aBlock] =
		      w->cubes.blockOfPosition[w->cubes.spacePosition + aBlock + 1];
		w->cubes.spacePosition = fillBeginPos;
		w->cubes.blockOfPosition[w->cubes.spacePosition] = 0;
	}	
	if ((dir == TOP) || (dir == BOTTOM)) {
		fillBeginPos = w->cubes.spacePosition +
		  numBlocks * w->cubes.sizeX;
		fillBeginI = Row(w, fillBeginPos) * w->cubes.offset.y +
		  w->cubes.delta.y + w->cubes.puzzleOffset.y + offsetPosition;
		if (numBlocks < 0)
		  numBlocks = -numBlocks;
		for (inc = 0; inc <= toMoveY;) {
		  for (aBlock = 0; aBlock < numBlocks; aBlock++) {
		    int dx, dy, dz, pos;
		    if (inc == 0) {
		      cb.reason = CUBES_MOVED;
		      XtCallCallbacks((Widget) w,
		        XtNselectCallback, &cb);
		      PutMove(dir);
		    }
		    if (dir == TOP)
		      pos = w->cubes.spacePosition +
		        (aBlock + 1) * w->cubes.sizeX;
		    else
		      pos = w->cubes.spacePosition -
		        (aBlock + 1) * w->cubes.sizeX;
		    if (w->cubes.vertical) {
		      dx = Column(w, pos) * w->cubes.offset.x +
		        w->cubes.delta.x + w->cubes.puzzleOffset.x +
		        offsetPosition;
		      dy = Row(w, pos) * w->cubes.offset.y +
		        w->cubes.delta.y + w->cubes.puzzleOffset.y +
		        offsetPosition;
		      dz = Stack(w, pos) * (w->cubes.sizeY *
		        w->cubes.offset.y + w->cubes.delta.y - 1);
		      dy += dz;
		      /* just add the first time */
		      if (aBlock == 0 && inc == 0)
		        fillBeginI += dz;
		    } else {
		      dx = Column(w, pos) * w->cubes.offset.x +
		        w->cubes.delta.x + w->cubes.puzzleOffset.x +
		        offsetPosition;
		      dy = Row(w, pos) * w->cubes.offset.y +
		        w->cubes.delta.y + w->cubes.puzzleOffset.y +
		        offsetPosition;
		      dz = Stack(w, pos) * (w->cubes.sizeX *
		        w->cubes.offset.x + w->cubes.delta.x - 1);
		      dx += dz;
		    }
		    {
		      int ix = dx;
		      int iy;

		      if (dir == TOP)
		        iy = dy - inc;
		      else /* dir == BOTTOM */
		        iy = dy + inc;
		      fill3DRect(w, w->cubes.blockGC,
		        w->cubes.blockDarkerGC, w->cubes.blockBrighterGC,
		        ix, iy, w->cubes.blockSize.x, w->cubes.blockSize.y,
		        True);
		      if (aBlock < numBlocks - 1) {
		       if (dir == TOP)
		          XFillRectangle(XtDisplay(w), XtWindow(w),
		            w->cubes.inverseGC,
		            ix, iy + w->cubes.blockSize.y,
		            w->cubes.blockSize.x, gapY);
		        else /* dir == BOTTOM */
		          XFillRectangle(XtDisplay(w), XtWindow(w),
		            w->cubes.inverseGC,
		            ix, iy - gapY, w->cubes.blockSize.x, gapY);
		      } else {
		        if (dir == TOP)
		          XFillRectangle(XtDisplay(w), XtWindow(w),
		            w->cubes.inverseGC,
		            ix, iy + w->cubes.blockSize.y,
		            w->cubes.blockSize.x, fillBeginI - iy);
		        else /* dir == BOTTOM */
		          XFillRectangle(XtDisplay(w), XtWindow(w),
		            w->cubes.inverseGC,
		            ix, fillBeginI,
		            w->cubes.blockSize.x, iy - fillBeginI);
		      }
		      {
		        /* draw block number after slide complete */
		        int i = 0, offsetX = 0;
		        int block = w->cubes.blockOfPosition[pos];
		        char buf[5];

		        (void) int2String(buf, block, w->cubes.base);
		        while (block >= 1) {
		          block /= w->cubes.base;
		          offsetX += w->cubes.digitOffset.x;
		          i++;
		        }
		        (void) XDrawString(XtDisplay(w), XtWindow(w),
		          w->cubes.borderGC,
			  ix + w->cubes.blockSize.x / 2 - offsetX,
			  iy + w->cubes.blockSize.y / 2 +
			  w->cubes.digitOffset.y, buf, i);
		      }
		    }
		  }
		  if (inc == toMoveY)
		    break;
		  if (toMoveY - inc > toMoveY / w->cubes.numSlices)
		    inc += toMoveY / w->cubes.numSlices;
		  else
		    inc = toMoveY;
		  XFlush(XtDisplay(w));
		  Sleep(w->cubes.moveDelay);
		}
		if (dir == TOP)
		  for (aBlock = 0; aBlock < numBlocks; aBlock++)
		    w->cubes.blockOfPosition[w->cubes.spacePosition + aBlock * w->cubes.sizeX] =
		      w->cubes.blockOfPosition[w->cubes.spacePosition + (aBlock + 1) * w->cubes.sizeX];
		else
		  for (aBlock = 0; aBlock < numBlocks; aBlock++)
		    w->cubes.blockOfPosition[w->cubes.spacePosition - aBlock * w->cubes.sizeX] =
		      w->cubes.blockOfPosition[w->cubes.spacePosition - (aBlock + 1) * w->cubes.sizeX];
		w->cubes.spacePosition = fillBeginPos;
		w->cubes.blockOfPosition[w->cubes.spacePosition] = 0;
	}
}

static void
AnimateMove(CubesWidget w, int fromPos)
{
	int pos = w->cubes.spacePosition, offsetPosition = 0;
	int dx, dy, dz, fromdx, fromdy, fromdz;
	int xToCenter, yToCenter;
	int incTo, incAt, inc;

	if (w->cubes.vertical) {
		dx = Column(w, pos) * w->cubes.offset.x + w->cubes.delta.x +
			w->cubes.puzzleOffset.x + offsetPosition;
		dy = Row(w, pos) * w->cubes.offset.y + w->cubes.delta.y +
			w->cubes.puzzleOffset.y + offsetPosition;
		dz = Stack(w, pos) * (w->cubes.sizeY * w->cubes.offset.y +
			w->cubes.delta.y - 1);
		fromdx = Column(w, fromPos) * w->cubes.offset.x + w->cubes.delta.x +
			w->cubes.puzzleOffset.x + offsetPosition;
		fromdy = Row(w, fromPos) * w->cubes.offset.y + w->cubes.delta.y +
			w->cubes.puzzleOffset.y + offsetPosition;
		fromdz = Stack(w, fromPos) * (w->cubes.sizeY * w->cubes.offset.y +
			w->cubes.delta.y - 1);
		fromdy += fromdz;
		dy += dz;
	} else {
		dx = Column(w, pos) * w->cubes.offset.x + w->cubes.delta.x +
			w->cubes.puzzleOffset.x + offsetPosition;
		dy = Row(w, pos) * w->cubes.offset.y + w->cubes.delta.y +
			w->cubes.puzzleOffset.y + offsetPosition;
		dz = Stack(w, pos) * (w->cubes.sizeX * w->cubes.offset.x +
			w->cubes.delta.x - 1);
		dx += dz;
		fromdx = Column(w, fromPos) * w->cubes.offset.x + w->cubes.delta.x +
			w->cubes.puzzleOffset.x + offsetPosition;
		fromdy = Row(w, fromPos) * w->cubes.offset.y + w->cubes.delta.y +
			w->cubes.puzzleOffset.y + offsetPosition;
		fromdz = Stack(w, fromPos) * (w->cubes.sizeX * w->cubes.offset.x +
			w->cubes.delta.x - 1);
		fromdx += fromdz;
	}
	xToCenter = dx + w->cubes.blockSize.x / 2;
	yToCenter = dy + w->cubes.blockSize.y / 2;
	incTo = MIN(w->cubes.blockSize.x, w->cubes.blockSize.y) / 2;
	incAt = MAX(incTo / (2 * w->cubes.numSlices), 1);

	XFillRectangle(XtDisplay(w), XtWindow(w),
		w->cubes.inverseGC,
		fromdx, fromdy, incAt, w->cubes.blockSize.y);
	XFillRectangle(XtDisplay(w), XtWindow(w),
		w->cubes.inverseGC,
		fromdx, fromdy, w->cubes.blockSize.x, incAt);
	XFillRectangle(XtDisplay(w), XtWindow(w),
		w->cubes.inverseGC,
		fromdx + w->cubes.blockSize.x - incAt, fromdy,
		incAt, w->cubes.blockSize.y);
	XFillRectangle(XtDisplay(w), XtWindow(w),
		w->cubes.inverseGC,
		fromdx, fromdy + w->cubes.blockSize.y - incAt,
		w->cubes.blockSize.x, incAt);
	for (inc = incAt; inc <= incTo; inc += incAt) {
		int xR = fromdx + inc, yR = fromdy + inc;
		int wR = w->cubes.blockSize.x - 2 * inc;
		int hR = w->cubes.blockSize.y - 2 * inc;

		/* start disappear block */
		if (wR > 0 && hR > 0) {
			XFillRectangle(XtDisplay(w), XtWindow(w),
				w->cubes.inverseGC,
				xR, yR, incAt, hR);
			XFillRectangle(XtDisplay(w), XtWindow(w),
				w->cubes.inverseGC,
				xR, yR, wR, incAt);
			XFillRectangle(XtDisplay(w), XtWindow(w),
				w->cubes.inverseGC,
				xR + wR - incAt, yR, incAt, hR);
			XFillRectangle(XtDisplay(w), XtWindow(w),
				w->cubes.inverseGC,
				xR, yR + hR - incAt, wR, incAt);
		}
		if (wR > 4 && hR > 4) {
			/* Perfectionism here...
			   have the erasing square be 3D also */
			XFillRectangle(XtDisplay(w), XtWindow(w),
				w->cubes.blockBrighterGC,
				xR + incAt, yR + 2, 2, hR - 4);
			XFillRectangle(XtDisplay(w), XtWindow(w),
				w->cubes.blockBrighterGC,
				xR + 2, yR + incAt, wR - 4, 2);
			XFillRectangle(XtDisplay(w), XtWindow(w),
				w->cubes.blockDarkerGC,
				xR + wR - incAt - 2, yR + 2, 2, hR - 4);
			XFillRectangle(XtDisplay(w), XtWindow(w),
				w->cubes.blockDarkerGC,
				xR + 2, yR + hR - incAt - 2, wR - 4, 2);
		}
		/* end disappear block */
		/* start appear blocks */
		if (w->cubes.blockSize.x > w->cubes.blockSize.y) {
			/* horizontal blocks */
			fill3DRect(w, w->cubes.blockGC,
			  w->cubes.blockDarkerGC,
			  w->cubes.blockBrighterGC,
			  xToCenter - inc - (w->cubes.blockSize.x / 2 - w->cubes.blockSize.y / 2),
			  yToCenter - inc,
			  2 * (inc + w->cubes.blockSize.x / 2 - w->cubes.blockSize.y / 2),
			  2 * inc,
			  True);
		} else {
			/* vertical blocks */
			fill3DRect(w, w->cubes.blockGC,
			  w->cubes.blockDarkerGC,
			  w->cubes.blockBrighterGC,
			  xToCenter - inc,
			  yToCenter - inc - (w->cubes.blockSize.y / 2 - w->cubes.blockSize.x / 2),
			  2 * inc,
			  2 * (inc + w->cubes.blockSize.y / 2 - w->cubes.blockSize.x / 2),
			  True);
		}
		XFlush(XtDisplay(w));
		Sleep(w->cubes.moveDelay);
	}
	XFillRectangle(XtDisplay(w), XtWindow(w),
		w->cubes.inverseGC,
		fromdx, fromdy, w->cubes.blockSize.x, w->cubes.blockSize.y);
	fill3DRect(w, w->cubes.blockGC, /* final draw to fix minor errors */
		w->cubes.blockDarkerGC,
		w->cubes.blockBrighterGC,
		dx, dy, w->cubes.blockSize.x, w->cubes.blockSize.y, True);
	w->cubes.spacePosition = fromPos;

	{
		int i = 0, offsetX = 0;
		int block = w->cubes.blockOfPosition[pos];
		char buf[5];

		(void) int2String(buf, block, w->cubes.base);
		while (block >= 1) {
			block /= w->cubes.base;
			offsetX += w->cubes.digitOffset.x;
			i++;
		}
		(void) XDrawString(XtDisplay(w), XtWindow(w), w->cubes.borderGC,
			dx + w->cubes.blockSize.x / 2 - offsetX,
			dy + w->cubes.blockSize.y / 2 +
			w->cubes.digitOffset.y, buf, i);
	}
}

static int
ToPosition(CubesWidget w, int column, int row, int stack)
{
	return (column + w->cubes.sizeX * row + w->cubes.sizeRect * stack);
}

static int
Column(CubesWidget w, int pos)
{
	return ((pos % w->cubes.sizeRect) % w->cubes.sizeX);
}

static int
Row(CubesWidget w, int pos)
{
	return ((pos % w->cubes.sizeRect) / w->cubes.sizeX);
}

static int
Stack(CubesWidget w, int pos)
{
	return (pos / w->cubes.sizeRect);
}

static int
int2String(char *buf, int number, int base)
{
	int         digit, mult = base, last, position;

	if (number < 0) {
		(void) printf("number %d < 0\n", number);
		return 0;
	}
	last = 1;
	while (number >= mult) {
		last++;
		mult *= base;
	}
	for (position = 0; position < last; position++) {
		mult /= base;
		digit = number / mult;
		number -= digit * mult;
		buf[position] = digit + '0';
		if (buf[position] > '9') {	/* ASCII */
			buf[position] += ('A' - '9' - 1);
		} else if (buf[position] < '0') {	/* EBCDIC */
			buf[position] += ('A' - '9' - 1);
			if (buf[position] > 'I')
				buf[position] += ('J' - 'I' - 1);
			if (buf[position] > 'R')
				buf[position] += ('S' - 'R' - 1);
		}
	}
	buf[last] = '\0';
	return last;
}

Boolean
CheckSolved(CubesWidget w)
{
	int         i;

	for (i = 1; i < w->cubes.sizeBlock; i++)
		if (w->cubes.blockOfPosition[i - 1] != i)
			return FALSE;
	return TRUE;
}
