/*-
# X-BASED PANEX(tm)
#
#  Panex.c
#
###
#
#  Copyright (c) 1996 - 2002	David Albert Bagley, bagleyd@tux.org
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/* Methods file for Panex */

#include <stdio.h>
#include <stdlib.h>
#ifdef VMS
#include <unixlib.h>
#else
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#endif
#include <X11/IntrinsicP.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/CoreP.h>
#include "PanexP.h"

#ifndef DATAFILE
#define DATAFILE "/usr/games/lib/panex.dat"
#endif

static void InitializePanex(Widget request, Widget renew);
static void ExposePanex(Widget renew, XEvent * event, Region region);
static void ResizePanex(PanexWidget w);
static void DestroyPanex(Widget old);
static Boolean SetValuesPanex(Widget current, Widget request, Widget renew);
static void QuitPanex(PanexWidget w, XEvent * event, char **args, int nArgs);
static void SelectPanex(PanexWidget w, XEvent * event, char **args, int nArgs);
static void ReleasePanex(PanexWidget w, XEvent * event, char **args, int nArgs);
static void GetPanex(PanexWidget w, XEvent * event, char **args, int nArgs);
static void WritePanex(PanexWidget w, XEvent * event, char **args, int nArgs);
static void ClearPanex(PanexWidget w, XEvent * event, char **args, int nArgs);
static void UndoPanex(PanexWidget w, XEvent * event, char **args, int nArgs);
static void SolvePanex(PanexWidget w, XEvent * event, char **args, int nArgs);
static void IncrementPanex(PanexWidget w, XEvent * event, char **args, int nArgs);
static void DecrementPanex(PanexWidget w, XEvent * event, char **args, int nArgs);
static void ModePanex(PanexWidget w, XEvent * event, char **args, int nArgs);
static void EnterPanex(PanexWidget w, XEvent * event, char **args, int nArgs);
static void LeavePanex(PanexWidget w, XEvent * event, char **args, int nArgs);

static int  SelectTile(PanexWidget w, int x);
static void SetAllColors(PanexWidget w, Boolean init);
static void GetColor(PanexWidget w, int pyramid, Boolean init);
static void CheckTiles(PanexWidget w);
static void GetTiles(PanexWidget w);
static void WriteTiles(PanexWidget w);
static void ClearTiles(PanexWidget w);
static void UndoTiles(PanexWidget w);
static void SolveTiles(PanexWidget w);
static void IncrementTiles(PanexWidget w);
static void DecrementTiles(PanexWidget w);
static void ModeTiles(PanexWidget w);
static void ResetTiles(PanexWidget w);
static void MoveNoTiles(PanexWidget w);
static int  MoveTile(PanexWidget w, int fromStack, int fromPosition, int toStack);
static void SlideTile(PanexWidget w, int fromStack, int fromPosition, int toStack, int toPosition);
static int  RequestMove(PanexWidget w, int fromStack, int fromPosition, int toStack);
static void DrawFrame(PanexWidget w, GC gc);
static void DrawSlots(PanexWidget w, GC gc);
static void DrawTile(PanexWidget w, int i, int j, Boolean erase, int offset);
static Pixel brighter(PanexWidget w, Pixel pixel);
static Pixel darker(PanexWidget w, Pixel pixel);
static void drawShadow(PanexWidget w, GC gc, int startX, int startY,
	int sizeX, int sizeY);
static void fill3DRect(PanexWidget w, GC gc, GC darkerGC, GC brighterGC,
	int x, int y, int width, int height, Boolean raised);
static void DrawPyramid(PanexWidget w, int color, int i, int j, int size, int offset);

#ifdef DEBUG
static void PrintStacks(PanexWidget w);
static void PrintTiles(PanexWidget w);

#endif

static char defaultTranslationsPanex[] =
"<KeyPress>q: Quit()\n\
   Ctrl<KeyPress>C: Quit()\n\
   <Btn1Down>: Select()\n\
   <Btn1Up>: Release()\n\
   <KeyPress>r: Clear()\n\
   <KeyPress>c: Clear()\n\
   <Btn3Down>(2+): Clear()\n\
   <KeyPress>g: Get()\n\
   <KeyPress>w: Write()\n\
   <KeyPress>u: Undo()\n\
   <KeyPress>s: Solve()\n\
   <KeyPress>i: Increment()\n\
   <KeyPress>d: Decrement()\n\
   <KeyPress>m: Mode()\n\
   <EnterWindow>: Enter()\n\
   <LeaveWindow>: Leave()";

static XtActionsRec actionsListPanex[] =
{
	{(char *) "Quit", (XtActionProc) QuitPanex},
	{(char *) "Select", (XtActionProc) SelectPanex},
	{(char *) "Release", (XtActionProc) ReleasePanex},
	{(char *) "Clear", (XtActionProc) ClearPanex},
	{(char *) "Get", (XtActionProc) GetPanex},
	{(char *) "Write", (XtActionProc) WritePanex},
	{(char *) "Undo", (XtActionProc) UndoPanex},
	{(char *) "Solve", (XtActionProc) SolvePanex},
	{(char *) "Increment", (XtActionProc) IncrementPanex},
	{(char *) "Decrement", (XtActionProc) DecrementPanex},
	{(char *) "Mode", (XtActionProc) ModePanex},
	{(char *) "Enter", (XtActionProc) EnterPanex},
	{(char *) "Leave", (XtActionProc) LeavePanex}
};

static XtResource resourcesPanex[] =
{
	{XtNuserName, XtCUserName, XtRString, sizeof (String),
	 XtOffset(PanexWidget, panex.username),
	 XtRString, (caddr_t) "nobody"},
	{XtNpyramidColor0, XtCLabel, XtRString, sizeof (String),
	 XtOffset(PanexWidget, panex.pyramidName[0]),
	 XtRString, (caddr_t) "Magenta"},
	{XtNpyramidColor1, XtCLabel, XtRString, sizeof (String),
	 XtOffset(PanexWidget, panex.pyramidName[1]),
	 XtRString, (caddr_t) "Yellow"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(PanexWidget, panex.foreground),
	 XtRString, (caddr_t) "Cyan" /*XtDefaultForeground*/},
	{XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(PanexWidget, panex.background),
	 XtRString, (caddr_t) "NavyBlue" /*XtDefaultBackground*/},
	{XtNtileColor, XtCColor, XtRPixel, sizeof (Pixel),
	 XtOffset(PanexWidget, panex.tileColor),
	 XtRString, (caddr_t) "LightGray" /*XtDefaultBackground*/},
	{XtNtileBorder, XtCColor, XtRPixel, sizeof (Pixel),
	 XtOffset(PanexWidget, panex.borderColor),
	 XtRString, (caddr_t) "gray25" /*XtDefaultForeground*/},
	{XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension),
	 XtOffset(PanexWidget, core.width),
	 XtRString, (caddr_t) "400"},
	{XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
	 XtOffset(PanexWidget, core.height),
	 XtRString, (caddr_t) "200"},
	{XtNtiles, XtCTiles, XtRInt, sizeof (int),
	 XtOffset(PanexWidget, panex.tiles),
	 XtRString, (caddr_t) "10"},	/*DEFAULTTILES */
	{XtNmode, XtCMode, XtRInt, sizeof (int),
	 XtOffset(PanexWidget, panex.mode),
	 XtRString, (caddr_t) "1"},	/*DEFAULTMODE */
	{XtNmono, XtCMono, XtRBoolean, sizeof (Boolean),
	 XtOffset(PanexWidget, panex.mono),
	 XtRString, (caddr_t) "FALSE"},
	{XtNreverse, XtCReverse, XtRBoolean, sizeof (Boolean),
	 XtOffset(PanexWidget, panex.reverse),
	 XtRString, (caddr_t) "FALSE"},
	{XtNdelay, XtCDelay, XtRInt, sizeof (int),
	 XtOffset(PanexWidget, panex.delay),
	 XtRString, (caddr_t) "300"},
	{XtNstart, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(PanexWidget, panex.started),
	 XtRString, (caddr_t) "TRUE"},
	{XtNmenu, XtCMenu, XtRInt, sizeof (int),
	 XtOffset(PanexWidget, panex.menu),
	 XtRString, (caddr_t) "-1"},
	{XtNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
	 XtOffset(PanexWidget, panex.select), XtRCallback, (caddr_t) NULL}
};

PanexClassRec panexClassRec =
{
	{
		(WidgetClass) & widgetClassRec,		/* superclass */
		(char *) "Panex",	/* class name */
		sizeof (PanexRec),	/* widget size */
		NULL,		/* class initialize */
		NULL,		/* class part initialize */
		FALSE,		/* class inited */
		(XtInitProc) InitializePanex,	/* initialize */
		NULL,		/* initialize hook */
		XtInheritRealize,	/* realize */
		actionsListPanex,	/* actions */
		XtNumber(actionsListPanex),	/* num actions */
		resourcesPanex,	/* resources */
		XtNumber(resourcesPanex),	/* num resources */
		NULLQUARK,	/* xrm class */
		TRUE,		/* compress motion */
		TRUE,		/* compress exposure */
		TRUE,		/* compress enterleave */
		TRUE,		/* visible interest */
		(XtWidgetProc) DestroyPanex,	/* destroy */
		(XtWidgetProc) ResizePanex,	/* resize */
		(XtExposeProc) ExposePanex,	/* expose */
		(XtSetValuesFunc) SetValuesPanex,	/* set values */
		NULL,		/* set values hook */
		XtInheritSetValuesAlmost,	/* set values almost */
		NULL,		/* get values hook */
		NULL,		/* accept focus */
		XtVersion,	/* version */
		NULL,		/* callback private */
		defaultTranslationsPanex,	/* tm table */
		NULL,		/* query geometry */
		NULL,		/* display accelerator */
		NULL		/* extension */
	},
	{
		0		/* ignore */
	}
};

WidgetClass panexWidgetClass = (WidgetClass) & panexClassRec;

static int  startPositions[MAXMODES][MAXSTACKS] =
{
	{0, -1, -1},
	{0, -1, 1}
};
static int  finishPositions[MAXMODES][MAXSTACKS] =
{
	{-1, -1, 0},
	{1, -1, 0}
};

static XPoint trapazoidUnit[5] =
{
	{0, 0},
	{1, 1},
	{-3, 0},
	{1, -1},
	{2, 0}
};

#ifndef HAVE_USLEEP
#if !defined( VMS ) || defined( XVMSUTILS ) ||  ( __VMS_VER >= 70000000 )
#ifdef USE_XVMSUTILS
#include <X11/unix_time.h>
#endif
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#if HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#endif
#endif
#if defined(SYSV) || defined(SVR4)
#ifdef LESS_THAN_AIX3_2
#include <sys/poll.h>
#else /* !LESS_THAN_AIX3_2 */
#include <poll.h>
#endif /* !LESS_THAN_AIX3_2 */
#endif /* defined(SYSV) || defined(SVR4) */

/* not static in case usleep found in system include */
int
usleep(unsigned int usec)
{
#if (defined (SYSV) || defined(SVR4)) && !defined(__hpux)
#if defined(HAVE_NANOSLEEP)
	{
		struct timespec rqt;

		rqt.tv_nsec = 1000 * (usec % (unsigned int) 1000000);
		rqt.tv_sec = usec / (unsigned int) 1000000;
		return nanosleep(&rqt, (struct timespec *) NULL);
	}
#else
	(void) poll(
#if defined(__cplusplus) || defined(c_plusplus)
		(pollfd *) /* guess */
#else
		(void *)
#endif
		0, (int) 0, (usec / 1000));	/* ms resolution */
#endif
#else
#ifdef VMS
	long        timadr[2];

	if (usec != 0) {
		timadr[0] = -usec * 10;
		timadr[1] = -1;

		sys$setimr(4, &timadr, 0, 0, 0);
		sys$waitfr(4);
	}
#else
	struct timeval time_out;

#if  0
	/* (!defined(AIXV3) && !defined(__hpux)) */
	extern int  select(int, fd_set *, fd_set *, fd_set *, struct timeval *);

#endif

	time_out.tv_usec = usec % (unsigned int) 1000000;
	time_out.tv_sec = usec / (unsigned int) 1000000;
	(void) select(0, (void *) 0, (void *) 0, (void *) 0, &time_out);
#endif
#endif
	return 0;
}
#endif

static void
Sleep(unsigned int cMilliseconds)
{
	(void) usleep(cMilliseconds * 1000);
}

static void
InitializePanex(Widget request, Widget renew)
{
	PanexWidget w = (PanexWidget) renew;
	int         stack;

	for (stack = 0; stack < MAXSTACKS; stack++) {
		w->panex.tileOfPosition[stack] = NULL;
		w->panex.positionOfTile[stack] = NULL;
	}
	CheckTiles(w);
	InitMoves();
	ResetTiles(w);
	w->panex.depth = DefaultDepthOfScreen(XtScreen(w));
	SetAllColors(w, True);
	ResizePanex(w);
}

static void
DestroyPanex(Widget old)
{
	PanexWidget w = (PanexWidget) old;
	int         pyramid;

	for (pyramid = 0; pyramid <= w->panex.mode; pyramid++)
		XtReleaseGC(old, w->panex.pyramidGC[pyramid]);
	XtReleaseGC(old, w->panex.puzzleGC);
	XtReleaseGC(old, w->panex.tileGC);
	XtReleaseGC(old, w->panex.borderGC);
	XtReleaseGC(old, w->panex.inverseGC);
	XtRemoveCallbacks(old, XtNselectCallback, w->panex.select);
}

static void
ResizePanex(PanexWidget w)
{
	w->panex.delta.x = 8;
	w->panex.delta.y = 2;
	w->panex.pos.x = MAX(((int) w->core.width - w->panex.delta.x) /
			     MAXSTACKS, w->panex.delta.x);
	w->panex.pos.y = MAX(((int) w->core.height - 2 * w->panex.delta.y - 3) /
			     (w->panex.tiles + 1), w->panex.delta.y);
	w->panex.width = w->panex.pos.x * MAXSTACKS + w->panex.delta.x + 2;
	w->panex.height = w->panex.pos.y * (w->panex.tiles + 1) +
		w->panex.delta.y + 3;
	w->panex.puzzleOffset.x = ((int) w->core.width - w->panex.width + 2) / 2;
	w->panex.puzzleOffset.y = ((int) w->core.height - w->panex.height + 2) / 2;
	/* Make the following odd */
	w->panex.tileSize.x = (((w->panex.pos.x - w->panex.delta.x) >> 1) << 1) + 1;
	w->panex.tileSize.y = w->panex.pos.y - w->panex.delta.y + 1;
	w->panex.letterOffset.x = 3;
	w->panex.letterOffset.y = 4;
}

static void
ExposePanex(Widget renew, XEvent * event, Region region)
{
	PanexWidget w = (PanexWidget) renew;

	if (w->core.visible) {
		XFillRectangle(XtDisplay(w), XtWindow(w), w->panex.inverseGC,
			0, 0, w->core.width, w->core.height);
		DrawFrame(w, (w->panex.focus) ? w->panex.puzzleGC :
			w->panex.inverseGC);
		DrawSlots(w, w->panex.borderGC);
		DrawAllTiles(w);
	}
}

static      Boolean
SetValuesPanex(Widget current, Widget request, Widget renew)
{
	PanexWidget c = (PanexWidget) current, w = (PanexWidget) renew;
	Boolean     redraw = False;
	Boolean     redrawTiles = False, setColors = False;
	int         pyramid;

	CheckTiles(w);
	for (pyramid = 0; pyramid <= w->panex.mode; pyramid++) {
		if (strcmp(w->panex.pyramidName[pyramid], c->panex.pyramidName[pyramid])) {
			setColors = True;
			break;
		}
	}
	if (w->panex.background != c->panex.background ||
	    w->panex.foreground != c->panex.foreground ||
	    w->panex.borderColor != c->panex.borderColor ||
	    w->panex.tileColor != c->panex.tileColor ||
	    w->panex.reverse != c->panex.reverse ||
	    w->panex.mono != c->panex.mono ||
	    setColors) {
		SetAllColors(w, False);
		redrawTiles = True;
	}
	if (w->panex.tiles != c->panex.tiles) {
		DrawFrame(c, c->panex.inverseGC);
		DrawSlots(c, c->panex.inverseGC);
		ResetTiles(w);
		ResizePanex(w);
		redraw = True;
	} else if (w->panex.mode != c->panex.mode) {
		ResetTiles(w);
		redraw = True;
	} else if (w->panex.puzzleOffset.x != c->panex.puzzleOffset.x ||
		   w->panex.puzzleOffset.y != c->panex.puzzleOffset.y) {
		DrawFrame(c, c->panex.inverseGC);
		DrawSlots(c, c->panex.inverseGC);
		ResizePanex(w);
		redraw = True;
	}
	if (w->panex.menu != -1) {
		switch (w->panex.menu) {
		case 0:
			w->panex.menu = -1;
			GetTiles(w);
			break;
		case 1:
			w->panex.menu = -1;
			WriteTiles(w);
			break;
		case 3:
			w->panex.menu = -1;
			ClearTiles(w);
			break;
		case 4:
			w->panex.menu = -1;
			UndoTiles(w);
			break;
		case 5:
			w->panex.menu = -1;
			SolveTiles(w);
			break;
		case 6:
			w->panex.menu = -1;
			IncrementTiles(w);
			break;
		case 7:
			w->panex.menu = -1;
			DecrementTiles(w);
			break;
		case 8:
			w->panex.menu = -1;
			ModeTiles(w);
			break;
		default:
			w->panex.menu = -1;
			break;
		}
	}
	if (redrawTiles && !redraw && XtIsRealized(renew) && renew->core.visible) {
		DrawSlots(c, c->panex.inverseGC);
		if (w->panex.focus)
			DrawFrame(w, w->panex.puzzleGC);
		DrawSlots(w, w->panex.borderGC);
		DrawAllTiles(w);
	}
	return (redraw);
}

static void
QuitPanex(PanexWidget w, XEvent * event, char **args, int nArgs)
{
	XtCloseDisplay(XtDisplay(w));
	exit(0);
}

static void
SelectPanex(PanexWidget w, XEvent * event, char **args, int nArgs)
{
	int         stack;

	if (CheckSolved(w)) {
		MoveNoTiles(w);
	} else if (((stack = SelectTile(w, event->xbutton.x)) >= 0) &&
	    ((w->panex.currentPosition = TopOfStack(w, stack)) >= 0)) {
		w->panex.currentStack = stack;
		DrawTile(w, stack, w->panex.currentPosition, False, TRUE);
	} else
		w->panex.currentStack = -1;
}

static void
ReleasePanex(PanexWidget w, XEvent * event, char **args, int nArgs)
{
	panexCallbackStruct cb;
	int         toStack, toPosition;

	if (w->panex.currentStack < 0)
		return;
	DrawTile(w, w->panex.currentStack, w->panex.currentPosition, True, TRUE);
	DrawTile(w, w->panex.currentStack, w->panex.currentPosition, False, FALSE);
	if ((toStack = SelectTile(w, event->xbutton.x)) >= 0 &&
	    toStack != w->panex.currentStack) {
		if ((toPosition = MovePanex(w,
					    w->panex.currentStack, w->panex.currentPosition, toStack)) >= 0) {
			if (CheckSolved(w)) {
				cb.reason = PANEX_SOLVED;
				XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
			}
		} else if (toPosition == -1) {
			cb.reason = PANEX_BLOCKED;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		} else {
			cb.reason = PANEX_ILLEGAL;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		}
	}
	w->panex.currentStack = -1;
}

static void
GetPanex(PanexWidget w, XEvent * event, char **args, int nArgs)
{
	GetTiles(w);
}

static void
WritePanex(PanexWidget w, XEvent * event, char **args, int nArgs)
{
	WriteTiles(w);
}

static void
ClearPanex(PanexWidget w, XEvent * event, char **args, int nArgs)
{
	ClearTiles(w);
}

static void
UndoPanex(PanexWidget w, XEvent * event, char **args, int nArgs)
{
	UndoTiles(w);
}

static void
SolvePanex(PanexWidget w, XEvent * event, char **args, int nArgs)
{
	SolveTiles(w);
}

static void
IncrementPanex(PanexWidget w, XEvent * event, char **args, int nArgs)
{
	IncrementTiles(w);
}

static void
DecrementPanex(PanexWidget w, XEvent * event, char **args, int nArgs)
{
	DecrementTiles(w);
}

static void
ModePanex(PanexWidget w, XEvent * event, char **args, int nArgs)
{
	ModeTiles(w);
}

static void
EnterPanex(PanexWidget w, XEvent * event, char **args, int nArgs)
{
	w->panex.focus = True;
	DrawFrame(w, w->panex.puzzleGC);
}

static void
LeavePanex(PanexWidget w, XEvent * event, char **args, int nArgs)
{
	w->panex.focus = False;
	DrawFrame(w, w->panex.inverseGC);
}

int
MovePanex(PanexWidget w, int fromStack, int fromPosition, int toStack)
{
	panexCallbackStruct cb;
	int         toPosition;

	if ((toPosition = MoveTile(w, fromStack, fromPosition, toStack)) >= 0) {
		cb.reason = PANEX_MOVED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		PutMove(fromStack, toStack);
	}
	return toPosition;
}

static int
SelectTile(PanexWidget w, int x)
{
	int         i;

	x -= w->panex.puzzleOffset.x;
	i = (x - w->panex.delta.x / 2) / w->panex.pos.x;
	if (i < 0)
		i = 0;
	else if (i >= MAXSTACKS)
		i = MAXSTACKS - 1;
	/*
	   y -= w->panex.puzzleOffset.y;
	   j = (y - w->panex.delta.y / 2) / w->panex.pos.y;
	   if (j < 0)
	   j = 0;
	   else if (j > w->panex.tiles)
	   j = w->panex.tiles;
	 */
	return i;
}


#define FACTOR 0.7
#define MAXINTENSITY 0xFFFF

static Pixel
brighter(PanexWidget w, Pixel pixel)
{
	XColor color;
	int i = (int) ((1 - FACTOR) * MAXINTENSITY);

	color.pixel = pixel;
	XQueryColor(XtDisplay(w), DefaultColormap(XtDisplay(w), XtWindow(w)),
		&color);
	if (color.red < i)
		color.red = i;
	if (color.green < i)
		color.green = i;
	if (color.blue < i)
		color.blue = i;
	color.red = (unsigned short) MIN(color.red / FACTOR, MAXINTENSITY);
	color.green = (unsigned short) MIN(color.green / FACTOR, MAXINTENSITY);
	color.blue = (unsigned short) MIN(color.blue / FACTOR, MAXINTENSITY);
	if (XAllocColor(XtDisplay(w),
			DefaultColormap(XtDisplay(w), XtWindow(w)),
			&color))
		return color.pixel;
	return pixel;
}

static Pixel
darker(PanexWidget w, Pixel pixel)
{
	XColor color;

	color.pixel = pixel;
	XQueryColor(XtDisplay(w), DefaultColormap(XtDisplay(w), XtWindow(w)),
		&color);
	color.red = (unsigned short) (color.red * FACTOR);
	color.green = (unsigned short) (color.green * FACTOR);
	color.blue = (unsigned short) (color.blue * FACTOR);
	if (XAllocColor(XtDisplay(w),
			DefaultColormap(XtDisplay(w), XtWindow(w)),
			&color))
		return color.pixel;
	return pixel;
}

static void
SetAllColors(PanexWidget w, Boolean init)
{
	XGCValues   values;
	XtGCMask    valueMask;
	int         pyramid;

	valueMask = GCForeground | GCBackground;
	if (w->panex.reverse) {
		values.background = w->panex.foreground;
		values.foreground = w->panex.background;
	} else {
		values.foreground = w->panex.background;
		values.background = w->panex.foreground;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->panex.inverseGC);
	w->panex.inverseGC = XtGetGC((Widget) w, valueMask, &values);
	values.background = w->panex.background;
	values.foreground = w->panex.foreground;
	if (!init)
		XtReleaseGC((Widget) w, w->panex.puzzleGC);
	w->panex.puzzleGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->panex.reverse) {
		if (w->panex.depth < 2 || w->panex.mono) {
			values.background = w->panex.foreground;
			values.foreground = w->panex.background;
		} else {
			values.background = w->panex.tileColor;
			values.foreground = w->panex.borderColor;
		}
	} else {
		if (w->panex.depth < 2 || w->panex.mono) {
			values.foreground = w->panex.foreground;
			values.background = w->panex.background;
		} else {
			values.foreground = w->panex.tileColor;
			values.background = w->panex.borderColor;
		}
	}
	if (!init)
		XtReleaseGC((Widget) w, w->panex.tileGC);
	w->panex.tileGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->panex.depth >= 2 && !w->panex.mono) {
		values.foreground = brighter(w, w->panex.tileColor);
	}
	if (!init)
		XtReleaseGC((Widget) w, w->panex.tileBrighterGC);
	w->panex.tileBrighterGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->panex.depth >= 2 && !w->panex.mono) {
		values.foreground = darker(w, w->panex.tileColor);
	}
	if (!init)
		XtReleaseGC((Widget) w, w->panex.tileDarkerGC);
	w->panex.tileDarkerGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->panex.reverse) {
		if (w->panex.depth < 2 || w->panex.mono) {
			values.foreground = w->panex.foreground;
			values.background = w->panex.background;
		} else {
			values.foreground = w->panex.tileColor;
			values.background = w->panex.borderColor;
		}
	} else {
		if (w->panex.depth < 2 || w->panex.mono) {
			values.background = w->panex.foreground;
			values.foreground = w->panex.background;
		} else {
			values.background = w->panex.tileColor;
			values.foreground = w->panex.borderColor;
		}
	}
	if (!init)
		XtReleaseGC((Widget) w, w->panex.borderGC);
	w->panex.borderGC = XtGetGC((Widget) w, valueMask, &values);
	for (pyramid = 0; pyramid <= w->panex.mode; pyramid++)
		GetColor(w, pyramid, init);

}

static void
GetColor(PanexWidget w, int pyramid, Boolean init)
{
	XGCValues   values;
	XtGCMask    valueMask;
	XColor      colorCell, rgb;

	valueMask = GCForeground | GCBackground;
	if (w->panex.reverse) {
		values.background = w->panex.foreground;
	} else {
		values.background = w->panex.background;
	}
	if (w->panex.depth > 1 && !w->panex.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				  DefaultColormap(XtDisplay(w), XtWindow(w)),
			  w->panex.pyramidName[pyramid], &colorCell, &rgb)) {
			values.foreground = w->panex.pyramidColor[pyramid] = colorCell.pixel;
			if (!init)
				XtReleaseGC((Widget) w, w->panex.pyramidGC[pyramid]);
			w->panex.pyramidGC[pyramid] = XtGetGC((Widget) w, valueMask, &values);
			return;
		} else {
			char        buf[121];

			(void) sprintf(buf, "Color name \"%s\" is not defined",
				       w->panex.pyramidName[pyramid]);
			XtWarning(buf);
		}
	}
	if (w->panex.reverse) {
		values.background = w->panex.foreground;
		values.foreground = w->panex.background;
	} else {
		values.background = w->panex.background;
		values.foreground = w->panex.foreground;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->panex.pyramidGC[pyramid]);
	w->panex.pyramidGC[pyramid] = XtGetGC((Widget) w, valueMask, &values);
}

static void
CheckTiles(PanexWidget w)
{
	char        buf[121];

	if (w->panex.tiles < 1) {
		(void) sprintf(buf, "Number of Tiles out of bounds, use 1..MAXINT");
		XtWarning(buf);
		w->panex.tiles = DEFAULTTILES;
	}
	if (w->panex.mode < HANOI || w->panex.mode > PANEX) {
		XtWarning("Mode is in error, use 0 for Hanoi, 1 for Panex");
		w->panex.mode = DEFAULTMODE;
	}
	if (w->panex.delay < 0) {
		(void) sprintf(buf, "Delay out of bounds, use 0..MAXINT");
		XtWarning(buf);
		w->panex.delay = -w->panex.delay;
	}
}

static void
GetTiles(PanexWidget w)
{
	FILE       *fp;
	int         c, i, tiles, mode, moves;
	panexCallbackStruct cb;

	if ((fp = fopen(DATAFILE, "r")) == NULL)
		(void) printf("Can not read %s for get.\n", DATAFILE);
	else {
		FlushMoves(w);
		while ((c = getc(fp)) != EOF && c != SYMBOL);
		(void) fscanf(fp, "%d", &tiles);
		if (tiles >= MINTILES) {
			for (i = w->panex.tiles; i < tiles; i++) {
				cb.reason = PANEX_INC;
				XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
			}
			for (i = w->panex.tiles; i > tiles; i--) {
				cb.reason = PANEX_DEC;
				XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
			}
		} else
			(void) printf("%s corrupted: tiles %d should be between %d and MAXINT\n",
				      DATAFILE, tiles, MINTILES);
		while ((c = getc(fp)) != EOF && c != SYMBOL);
		(void) fscanf(fp, "%d", &mode);
		if (mode >= HANOI && mode <= PANEX) {
			if (w->panex.mode != mode) {
				cb.reason = PANEX_MODE;
				XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
			}
		} else
			(void) printf("%s corrupted: mode %d should be between %d and %d\n",
				      DATAFILE, mode, HANOI, PANEX);
		while ((c = getc(fp)) != EOF && c != SYMBOL);
		(void) fscanf(fp, "%d", &moves);
		ScanStartPosition(fp, w);
		cb.reason = PANEX_RESTORE;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		SetStartPosition(w);
		ScanMoves(fp, w, moves);
		(void) fclose(fp);
		if (mode == 0)
			(void) printf("%s: mode hanoi, tiles %d, moves %d.\n",
				      DATAFILE, tiles, moves);
		else
			(void) printf("%s: mode panex, tiles %d, moves %d.\n",
				      DATAFILE, tiles, moves);
	}
}

static void
WriteTiles(PanexWidget w)
{
	FILE       *fp;

	if ((fp = fopen(DATAFILE, "w")) == NULL)
		(void) printf("Can not write to %s.\n", DATAFILE);
	else {
		(void) fprintf(fp, "tiles%c %d\n", SYMBOL, w->panex.tiles);
		(void) fprintf(fp, "mode%c %d\n", SYMBOL, w->panex.mode);
		(void) fprintf(fp, "moves%c %d\n", SYMBOL, NumMoves());
		PrintStartPosition(fp, w);
		PrintMoves(fp);
		(void) fclose(fp);
		(void) printf("Saved to %s.\n", DATAFILE);
	}
}

static void
ClearTiles(PanexWidget w)
{
	panexCallbackStruct cb;

	DrawSlots(w, w->panex.inverseGC);
	ResetTiles(w);
	DrawSlots(w, w->panex.borderGC);
	DrawAllTiles(w);
	cb.reason = PANEX_RESET;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static void
UndoTiles(PanexWidget w)
{
	if (MadeMoves()) {
		panexCallbackStruct cb;
		int         fromStack, fromPosition, toStack;

		GetMove(&toStack, &fromStack);
		if ((fromPosition = TopOfStack(w, fromStack)) < 0 ||
		    MoveTile(w, fromStack, fromPosition, toStack) < 0)
			(void) printf(
					     "move from %d to %d can not be made.", fromStack, toStack);
		else {
			cb.reason = PANEX_UNDO;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		}
	}
}

static void
SolveTiles(PanexWidget w)
{
	panexCallbackStruct cb;

	/* Cheat and Reset To Start */
	DrawSlots(w, w->panex.inverseGC);
	ResetTiles(w);
	DrawSlots(w, w->panex.borderGC);
	DrawAllTiles(w);
	cb.reason = PANEX_RESET;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	SolveTilesFromStart(w);
}

static void
IncrementTiles(PanexWidget w)
{
	panexCallbackStruct cb;

	cb.reason = PANEX_INC;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static void
DecrementTiles(PanexWidget w)
{
	panexCallbackStruct cb;

	if (w->panex.tiles <= MINTILES)
		return;
	cb.reason = PANEX_DEC;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static void
ModeTiles(PanexWidget w)
{
	panexCallbackStruct cb;

	cb.reason = PANEX_MODE;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static void
ResetTiles(PanexWidget w)
{
	int         stack, loc;
	PanexLoc    i;

	w->panex.currentStack = -1;
	for (stack = 0; stack < MAXMODES; stack++) {
		if (w->panex.positionOfTile[stack]) {
			free(w->panex.positionOfTile[stack]);
			w->panex.positionOfTile[stack] = NULL;
		}
		if (startLoc[stack]) {
			free(startLoc[stack]);
			startLoc[stack] = NULL;
		}
	}
	for (stack = 0; stack <= w->panex.mode; stack++) {
		if (!(w->panex.positionOfTile[stack] = (PanexLoc *)
		      malloc(sizeof (PanexLoc) * w->panex.tiles)))
			XtError("Not enough memory, exiting.");
		if (!(startLoc[stack] = (PanexLoc *)
		      malloc(sizeof (PanexLoc) * w->panex.tiles)))
			XtError("Not enough memory, exiting.");
	}
	for (stack = 0; stack < MAXSTACKS; stack++) {
		if (w->panex.tileOfPosition[stack])
			free(w->panex.tileOfPosition[stack]);
		if (!(w->panex.tileOfPosition[stack] = (PanexLoc *)
		      malloc(sizeof (PanexLoc) * (w->panex.tiles + 1))))
			XtError("Not enough memory, exiting.");
	}
	for (stack = 0; stack <= w->panex.mode; stack++) {
		i.stack = 2 * stack;
		for (loc = 0; loc < w->panex.tiles; loc++) {
			i.loc = loc + 1;
			w->panex.positionOfTile[stack][loc] = i;
		}
	}
	for (stack = 0; stack < MAXSTACKS; stack++) {
		i.stack = -1;
		i.loc = -1;
		w->panex.tileOfPosition[stack][0] = i;
		if ((i.stack = startPositions[w->panex.mode][stack]) >= 0)
			for (loc = 1; loc <= w->panex.tiles; loc++) {
				i.loc = loc - 1;
				w->panex.tileOfPosition[stack][loc] = i;
		} else
			for (loc = 1; loc <= w->panex.tiles; loc++)
				w->panex.tileOfPosition[stack][loc] = i;
	}
	FlushMoves(w);
	w->panex.started = False;
}

static void
MoveNoTiles(PanexWidget w)
{
	panexCallbackStruct cb;

	cb.reason = PANEX_IGNORE;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

int
TopOfStack(PanexWidget w, int stack)
{
	int         i;

	for (i = 0; i <= w->panex.tiles; i++)
		if (w->panex.tileOfPosition[stack][i].stack >= 0)
			return i;
	return -1;
}

static int
RequestMove(PanexWidget w, int fromStack, int fromPosition, int toStack)
{
	int         i;

	/* Do not have to check above stack since it is the top one */
	if (toStack > fromStack)
		for (i = fromStack + 1; i <= toStack; i++) {
			if (w->panex.tileOfPosition[i][0].stack >= 0)
				return (-1);
	} else			/* Already ruled out toStack == fromStack */
		for (i = fromStack - 1; i >= toStack; i--) {
			if (w->panex.tileOfPosition[i][0].stack >= 0)
				return (-1);
		}
	i = TopOfStack(w, toStack);
	i = (i == -1) ? w->panex.tiles : i - 1;
	if (w->panex.mode == HANOI) {
		if (i == w->panex.tiles || w->panex.tileOfPosition[toStack][i + 1].loc >
		    w->panex.tileOfPosition[fromStack][fromPosition].loc)
			return i;
		else
			return -2;
	} else {
		if (i > w->panex.tileOfPosition[fromStack][fromPosition].loc + 1)
			return (w->panex.tileOfPosition[fromStack][fromPosition].loc + 1);
		else
			return i;
	}
}

static int
MoveTile(PanexWidget w, int fromStack, int fromPosition, int toStack)
{
	int         toPosition;

	if ((toPosition = RequestMove(w, fromStack, fromPosition, toStack)) >= 0)
		SlideTile(w, fromStack, fromPosition, toStack, toPosition);
	return toPosition;
}

static void
SlideTile(PanexWidget w, int fromStack, int fromPosition, int toStack, int toPosition)
{
	PanexLoc    top;
	int         currentStack = fromStack, currentPosition = fromPosition;
	int         nextStack = fromStack, nextPosition = fromPosition;

	while (currentPosition > 0) {
		DrawTile(w, currentStack, currentPosition, True, FALSE);
		top = w->panex.tileOfPosition[currentStack][currentPosition];
		nextPosition = currentPosition - 1;
		w->panex.tileOfPosition[nextStack][nextPosition] = top;
		w->panex.positionOfTile[top.stack][top.loc].stack = nextStack;
		w->panex.positionOfTile[top.stack][top.loc].loc = nextPosition;
		w->panex.tileOfPosition[currentStack][currentPosition].stack = -1;
		currentPosition = nextPosition;
		DrawTile(w, currentStack, currentPosition, False, FALSE);
		XFlush(XtDisplay(w));
		Sleep((unsigned int) w->panex.delay / (w->panex.tiles + MAXSTACKS - 1));
	}
	while (currentStack != toStack) {
		DrawTile(w, currentStack, currentPosition, True, FALSE);
		top = w->panex.tileOfPosition[currentStack][currentPosition];
		nextStack = (currentStack < toStack) ? currentStack + 1 :
			currentStack - 1;
		w->panex.tileOfPosition[nextStack][nextPosition] = top;
		w->panex.positionOfTile[top.stack][top.loc].stack = nextStack;
		w->panex.positionOfTile[top.stack][top.loc].loc = nextPosition;
		w->panex.tileOfPosition[currentStack][currentPosition].stack = -1;
		currentStack = nextStack;
		DrawTile(w, currentStack, currentPosition, False, FALSE);
		XFlush(XtDisplay(w));
		Sleep((unsigned int) w->panex.delay / (w->panex.tiles + MAXSTACKS - 1));
	}
	while (currentPosition < toPosition) {
		DrawTile(w, currentStack, currentPosition, True, FALSE);
		top = w->panex.tileOfPosition[currentStack][currentPosition];
		nextPosition = currentPosition + 1;
		w->panex.tileOfPosition[nextStack][nextPosition] = top;
		w->panex.positionOfTile[top.stack][top.loc].stack = nextStack;
		w->panex.positionOfTile[top.stack][top.loc].loc = nextPosition;
		w->panex.tileOfPosition[currentStack][currentPosition].stack = -1;
		currentPosition = nextPosition;
		DrawTile(w, currentStack, currentPosition, False, FALSE);
		XFlush(XtDisplay(w));
		Sleep((unsigned int) w->panex.delay / (w->panex.tiles + MAXSTACKS - 1));
	}
}

static void
DrawFrame(PanexWidget w, GC gc)
{
	XDrawRectangle(XtDisplay(w), XtWindow(w), gc,
	     0, 0, w->core.width - 1, w->core.height - 1);
}

static void
DrawSlots(PanexWidget w, GC gc)
{
	int         i, dx, dy, y, sizey;

	sizey = w->panex.tileSize.y / 3;
	if (sizey == 0)
		sizey = 1;
	dx = w->panex.tileSize.x / 3 + w->panex.delta.x + w->panex.puzzleOffset.x;
	y = (w->panex.tiles + 1) * w->panex.pos.y + w->panex.delta.y;
	dy = w->panex.delta.y / 2 + w->panex.puzzleOffset.y + sizey;
	if (gc == w->panex.inverseGC) {
		XFillRectangle(XtDisplay(w), XtWindow(w), gc,
			1, 1, w->core.width - 1, w->core.height - 1);
		return;
	}

	XFillRectangle(XtDisplay(w), XtWindow(w), gc, dx, dy,
	  w->panex.tileSize.x / 3 + (MAXSTACKS - 1) * w->panex.pos.x,
		       w->panex.tileSize.y / 3);
	for (i = 0; i < MAXSTACKS; i++) {
		XFillRectangle(XtDisplay(w), XtWindow(w), gc, dx, dy,
		w->panex.tileSize.x / 3, y - 1 - 2 * w->panex.tileSize.y / 3);
		dx += w->panex.pos.x;
	}
}

void
DrawAllTiles(PanexWidget w)
{
	int         i, j;

	for (i = 0; i < MAXSTACKS; i++)
		for (j = 0; j <= w->panex.tiles; j++)
			if (w->panex.tileOfPosition[i][j].stack >= 0)
				DrawTile(w, i, j, False, FALSE);
}

static void
drawShadow(PanexWidget w, GC gc, int startX, int startY,
		int sizeX, int sizeY)
{
	XFillRectangle(XtDisplay(w), XtWindow(w), gc,
		       startX, startY, sizeX, 1);
	XFillRectangle(XtDisplay(w), XtWindow(w), gc,
		       startX, startY, 1, sizeY);
}

static void
fill3DRect(PanexWidget w, GC gc, GC darkerGC, GC brighterGC,
		int x, int y, int width, int height, Boolean raised)
{
	GC currentGC = (raised) ? gc : darkerGC;

	if (width > 2 && height > 2)
		XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
			x + 1, y + 1, width - 2, height - 2);
	currentGC = (raised) ? brighterGC : darkerGC;
	XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
		x, y, 1, height);
	if (width > 1)
		XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
			x + 1, y, width - 1, 1);
	XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
		x + 1, y, 1, height);
	if (width > 2)
		XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
			x + 2, y + 1, width - 2, 1);
	currentGC = (raised) ? darkerGC : gc;
	if (width > 1 && height > 1)
		XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
			x + 1, y + height - 1, width - 1, 1);
	if (width > 1 && height > 1)
		XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
			x + width - 1, y, 1, height - 1);
	if (width > 3 && height > 2)
		XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
			x + 2, y + height - 2, width - 3, 1);
	if (width > 2 && height > 3)
		XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
			x + width - 2, y + 1, 1, height - 3);
}
static void
DrawTile(PanexWidget w, int i, int j, Boolean erase, int offsetPosition)
{
	int         dx, dy;
	GC          tileGC, borderGC;

	if (erase) {
		tileGC = w->panex.inverseGC;
		borderGC = w->panex.borderGC;
	} else {
		tileGC = w->panex.tileGC;
		borderGC = w->panex.borderGC;
	}
	dx = i * w->panex.pos.x + w->panex.delta.x + w->panex.puzzleOffset.x;
	dy = j * w->panex.pos.y + w->panex.delta.y + w->panex.puzzleOffset.y;
	if (erase)  {
		/* Draw Slots */
		XFillRectangle(XtDisplay(w), XtWindow(w), tileGC,
			dx + offsetPosition, dy + offsetPosition,
			w->panex.tileSize.x, w->panex.tileSize.y);
		if (j == 0) {
			if (i == 0) {
				XFillRectangle(XtDisplay(w), XtWindow(w), borderGC,
					dx + w->panex.tileSize.x / 3 + offsetPosition,
					dy + w->panex.tileSize.y / 3 - 1,
					2 * w->panex.tileSize.x / 3 + 2,
					w->panex.tileSize.y / 3);
			} else if (i == MAXSTACKS - 1) {
				XFillRectangle(XtDisplay(w), XtWindow(w), borderGC,
					dx + offsetPosition,
					dy + w->panex.tileSize.y / 3 - 1,
					2 * w->panex.tileSize.x / 3 - 1,
					w->panex.tileSize.y / 3 + offsetPosition);
			} else {
				XFillRectangle(XtDisplay(w), XtWindow(w), borderGC,
					dx + offsetPosition,
					dy + w->panex.tileSize.y / 3 - 1,
					w->panex.tileSize.x + 1,
					w->panex.tileSize.y / 3);
			}
			XFillRectangle(XtDisplay(w), XtWindow(w), borderGC,
				dx + w->panex.tileSize.x / 3,
				dy + w->panex.tileSize.y / 3 - 1,
				w->panex.tileSize.x / 3,
				w->panex.tileSize.y - w->panex.tileSize.y / 3 + 2 + offsetPosition - 1);
		} else if (j == w->panex.tiles)
			XFillRectangle(XtDisplay(w), XtWindow(w), borderGC,
				dx + w->panex.tileSize.x / 3, dy,
				w->panex.tileSize.x / 3,
				2 * w->panex.tileSize.y / 3 + 2);
		else
			XFillRectangle(XtDisplay(w), XtWindow(w), borderGC,
				dx + w->panex.tileSize.x / 3, dy,
				w->panex.tileSize.x / 3,
				w->panex.tileSize.y + offsetPosition);
	} else {
		if (offsetPosition != 0) {
			drawShadow(w, w->panex.tileDarkerGC,
				dx, dy, w->panex.tileSize.x, w->panex.tileSize.y);
		}
		dx += offsetPosition;
		dy += offsetPosition;

		fill3DRect(w, tileGC,
			w->panex.tileDarkerGC, w->panex.tileBrighterGC,
			dx, dy, w->panex.tileSize.x, w->panex.tileSize.y,
			offsetPosition == 0);
		DrawPyramid(w,
			    w->panex.tileOfPosition[i][j].stack, i, j,
			    w->panex.tileOfPosition[i][j].loc, offsetPosition);
	}
}

static void
DrawPyramid(PanexWidget w, int color, int i, int j, int size, int offsetPosition)
{
	XPoint      trapazoidList[5];
	int         k;
	GC          pyramidGC;

	for (k = 0; k <= 4; k++) {
		if (ABS(trapazoidUnit[k].x) == 3)
			trapazoidList[k].x = (size + 1) * SIGN(trapazoidUnit[k].x) *
				w->panex.tileSize.x / (w->panex.tiles + 1);
		else if (ABS(trapazoidUnit[k].x) == 2)
			trapazoidList[k].x = size * SIGN(trapazoidUnit[k].x) *
				w->panex.tileSize.x / (w->panex.tiles + 1);
		else
			trapazoidList[k].x = w->panex.tileSize.x /
				(2 * (w->panex.tiles + 1));
		trapazoidList[k].y = (w->panex.tileSize.y - 4) * trapazoidUnit[k].y;
	}
	k = w->panex.delta.x + i * w->panex.pos.x + w->panex.puzzleOffset.x +
		w->panex.tileSize.x / 2 + offsetPosition + 1;
	trapazoidList[0].x = trapazoidList[4].x / 2 + k;
	trapazoidList[0].y = j * w->panex.pos.y + w->panex.delta.y +
		w->panex.puzzleOffset.y + offsetPosition + 2;
	if (w->panex.mono || w->panex.depth < 2) {
		pyramidGC = (offsetPosition) ? w->panex.tileGC : w->panex.borderGC;
	} else {
		if ((w->panex.pyramidColor[color] == w->panex.borderColor) && offsetPosition)
			pyramidGC = w->panex.tileGC;
		else if ((w->panex.pyramidColor[color] == w->panex.tileColor) && !offsetPosition)
			pyramidGC = w->panex.borderGC;
		else
			pyramidGC = w->panex.pyramidGC[color];

	}
	XFillPolygon(XtDisplay(w), XtWindow(w), pyramidGC,
		     trapazoidList, 4, Convex, CoordModePrevious);
	if (w->panex.mono || w->panex.depth < 2) {

		char        buf[2];

		(void) sprintf(buf, "%c", w->panex.pyramidName[color][0]);
		(void) XDrawString(XtDisplay(w), XtWindow(w),
			    (offsetPosition) ? w->panex.borderGC : w->panex.tileGC,
			    k - w->panex.letterOffset.x, trapazoidList[0].y +
		  w->panex.tileSize.y / 2 + w->panex.letterOffset.y, buf, 1);
	}
}

#if 0
/* Only applies to Panex */
static int  middlePositions[MAXSTACKS] =
{-1, 0, 1};

Boolean
CheckMiddle(PanexWidget w)
{
	int         stack, loc;
	PanexLoc    i;

	for (stack = 0; stack < MAXSTACKS; stack++)
		if ((i.stack = middlePositions[stack]) >= 0)
			for (loc = 1; loc <= w->panex.tiles; loc++) {
				i.loc = loc - 1;
				if (w->panex.tileOfPosition[stack][loc].stack != i.stack ||
				    w->panex.tileOfPosition[stack][loc].loc != i.loc)
					return False;
			}
	return True;
}
#endif

Boolean
CheckSolved(PanexWidget w)
{
	int         stack, loc;
	PanexLoc    i;

	for (stack = 0; stack < MAXSTACKS; stack++)
		if ((i.stack = finishPositions[w->panex.mode][stack]) >= 0)
			for (loc = 1; loc <= w->panex.tiles; loc++) {
				i.loc = loc - 1;
				if (w->panex.tileOfPosition[stack][loc].stack != i.stack ||
				    w->panex.tileOfPosition[stack][loc].loc != i.loc)
					return False;
			}
	return True;
}

#ifdef DEBUG
void
PrintStacks(PanexWidget w)
{
	int         stack, position;

	(void) printf("top: where are the tiles in the stack\n");
	for (position = 0; position <= w->panex.tiles; position++)
		for (stack = 0; stack < MAXSTACKS; stack++) {
			printf("%d,%d", w->panex.tileOfPosition[stack][position].stack,
			       w->panex.tileOfPosition[stack][position].loc);
			if (stack + 1 == MAXSTACKS)
				(void) printf("\n");
			else
				(void) printf(" | ");
		}
}

void
PrintTiles(PanexWidget w)
{
	int         stack, position;

	(void) printf("pot: which stack are the tiles in\n");
	for (position = 0; position < w->panex.tiles; position++)
		for (stack = 0; stack <= w->panex.mode; stack++) {
			printf("%d,%d", w->panex.positionOfTile[stack][position].stack,
			       w->panex.positionOfTile[stack][position].loc);
			if (stack == w->panex.mode)
				(void) printf("\n");
			else
				(void) printf(" | ");
		}
}

#endif
